
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.PriorityQueue;
import java.util.Random;
import java.util.Set;
import java.util.Stack;
import java.util.TreeMap;

import edu.uci.ics.jung.graph.UndirectedSparseGraph;
//import edu.uci.ics.jung.algorithms.scoring.BetweennessCentrality;
import edu.uci.ics.jung.algorithms.importance.BetweennessCentrality;
import edu.uci.ics.jung.algorithms.scoring.DegreeScorer;


/*!
 * Contains the number of instances of each knowledge field.
 * Contains a method to compute betweeness centrality.
 * 
 *  \author Sirio Capizzi
 *  \author Martina Casari
 *  \author Guido Fioretti
 *  \author Ruggero Rossi
 *  \copyright GNU Public License.
 */
public class CommonKnowledge {
	
	/** Common knowledge is made of items composed by a knowledge field, the 
	 * number of instances there are of it, and the performance it provides. 
	 * It is organized as a hashmap with Key: KnowledgeField, and 
	 * Value: CommonKnowledgeData, where CommonKnowledgeData is composed of the 
	 * number of instances of this knowledge field and its performance.
	 */
	private HashMap<KnowledgeField, CommonKnowledgeData> knowledgeFields = 
			new HashMap<KnowledgeField, CommonKnowledgeData>();
	
	/** A boolean variable telling whether the above map changed.
	 * It is set to 'true' by updateField and it is set to 'false' by updatePerformance. 
	 */
	private boolean mapsChanged;
	
	/** 
	 * Gets the whole tree of knowledge fields.
	 * 
	 * @return Return the whole tree.
	 */
	public Map<KnowledgeField, CommonKnowledgeData> getKnowledgeFields() {
		return knowledgeFields;
	}	
	
	/**
	 * It updates the number of instances of a knowledge field.
	 * It takes care that the calling knowledge field is eventually renamed as 
	 * an instance of another knowledge field.
	 * This method works with delta_n <= 1, but it is called with delta_n = +-1.
	 * With delta_n = 1 the calling knowledge field is eventually renamed.
	 * (ONLY the calling knowledge field is renamed, that's why it doesn't work 
	 * with delta_n > 1). If delta_n < 0 renaming is not necessary because the 
	 * calling field(s) will be destroyed. In this casef it is asked to update to a 
	 * non-positive number of instances, it deletes the knowledge field.
	 * 
	 * @param kf
	 * @param delta_n
	 */
	public void updateField(KnowledgeField kf, int delta_n) {
//System.out.println("dentro updateField: delta n = " + delta_n);
		if (kf != null && delta_n != 0) {
			// If it is already there, get it and either add it or delete it 
			// from common knowledge.
//System.out.println(kf);
			if (this.containsKF(kf)) {  // calls method below
				KnowledgeField s = this.getAlterEgo(kf);
//System.out.println(s);
				// the new number of instances of kf
				int nNew = knowledgeFields.get(s).KFIs + delta_n;
//System.out.println(knowledgeFields.get(s).KFIs + delta_n);
				// stores the performance of s
				double p = knowledgeFields.get(s).performance;  
				// If the new number of instances of this knowledge is still 
				// positive, update it.
				if (nNew > 0) {
					// the number of instances of kf has been updated
					knowledgeFields.put(s, new CommonKnowledgeData(nNew, p));  
//System.out.println("put  stato attivato");
//System.out.println(this.getKnowledgeFields());
					mapsChanged = true;
					// If the proposed new knowledge field is equal to an 
					// existing one, change its ID.
					if (delta_n > 0)  // if kf has been added
						// only this kf's ID is changed!
						kf.changeID(s.getID()); 
				}
				// If the new number of instances of this knowledge field is 
				// zero, delete it.
				else if (nNew <= 0) {
					knowledgeFields.remove(s);
					mapsChanged = true;
				}
			}
			// If it's not already there and it is asked to add it to 
			// 'knowledgeFields', add it.
			else if (delta_n > 0) {
				// provisional value of performance: -0.1
//System.out.println("put  stato attivato");
				knowledgeFields.put(kf, new CommonKnowledgeData(delta_n, -0.1));  
				mapsChanged = true;
			}
		}
	}

	/**
	 * Gets the number of instances of a given field. Called by Model.
	 * If the field is not there, it returns zero.
	 *
	 * @param kf
	 * @return Return the instances of a given field. If the field is not there, 
	 * it returns zero.
	 */
	public int getNumInstancesOfField(KnowledgeField kf) {
		Integer n = 0;
		for (KnowledgeField j : knowledgeFields.keySet()) {
			if (j.getID() == kf.getID())
				n = knowledgeFields.get(j).KFIs;
		}
		return n;
	}
	
	/**
	 * Tells whether the map 'knowledgeFields' contains a specific knowledge 
	 * field. One cannot rely on the in-built method 'containsKey()' because it 
	 * is based on knowledge fields hashcodes. We want to check whether a 
	 * knowledge field is there, which has the same BEIs of another knowledge 
	 * field. If it does, its ID must be changed because it must be the same 
	 * knowledge field for us.
	 *
	 * @param kf
	 * @return Return true if the map 'knowledgeFields' contains a specific
	 *  knowledge field, false otherwise.
	 */
	public boolean containsKF(KnowledgeField kf) {
		boolean contains = false;
		// Compares kf with the knowledge fields recorded in Common Knowledge.
		for (KnowledgeField j : knowledgeFields.keySet()) {
			boolean sameBEIs = true;
			// Scans all the pairs of BEIs of knowledge fields j and kf.
			for (int i=0; i<Model.numBusinessElements; i++) {
				if (j.getInstance(i).getInstanceOfAspect() != kf.getInstance(i).getInstanceOfAspect())  
					// BEIs' IDs
					sameBEIs = false;
			}
			if (sameBEIs == true)
				contains = true;
		}
		return contains;		
	}
	
	/**
	 * Same procedure as 'containsKey' above, but the knowledge field is 
	 * returned, which has the same BEIs as kf albeit (for the moment) a 
	 * different ID.
	 *
	 * @param kf
	 * @return Return the knowledge field
	 */
	private KnowledgeField getAlterEgo(KnowledgeField kf) {
		KnowledgeField sosia = null;
		// Compares kf with the knowledge fields recorded in Common Knowledge.
		for (KnowledgeField j : knowledgeFields.keySet()) {
			boolean sameBEIs = true;
			// Scans all the pairs of BEIs of knowledge fields j and kf.
			for (int i=0; i<Model.numBusinessElements; i++) {
				if (j.getInstance(i).getInstanceOfAspect() != kf.getInstance(i).getInstanceOfAspect())
					sameBEIs = false;
			}
			if (sameBEIs == true)
				sosia = j;
		}
		return sosia;
	}

	/**
	 * CommonKnowledge may have stored a different knowledge field, entailing 
	 * the same BEIs. QUESTO DEVE ESSERE EVITATO
	 *
	 * @param kf
	 * @return Returns the performance of a particular knowledge field.
	 */
	public double getPerformance(KnowledgeField kf) {
		// The equivalent knowledge field that is actually stored in 
		// CommonKnowledge.
		KnowledgeField equivalentKF = kf;
		// Compares kf with the knowledge fields recorded in Common Knowledge.
		for (KnowledgeField j : knowledgeFields.keySet()) {
			boolean sameBEIs = true;
			// Scans all the pairs of BEIs of knowledge fields j and kf.
			for (int i=0; i<Model.numBusinessElements; i++) {
				if (j.getInstance(i).getInstanceOfAspect() != kf.getInstance(i).getInstanceOfAspect())  
					// BEIs' IDs
					sameBEIs = false;
			}
			if (sameBEIs == true)
				equivalentKF = j;
		}
		return knowledgeFields.get(equivalentKF).performance;
	}

	/**
	 * Updates performance if the set of knowledge fields changed.
	 * Performance is computed according to the algorithm specified by 
	 * originOfPerformance.
	 */
	public void updatePerformance() {
		if (mapsChanged == true) {
			if (Model.originOfPerformance == 0)
				computeRandomPerformance();
			if (Model.originOfPerformance == 1)
				computeBetweenness();
			if (Model.originOfPerformance == 2)
				computeDegreeCentrality();
			mapsChanged = false;
		}
	}
	
	/**
	 * Computes random performance.
	 */
	private void computeRandomPerformance() {
		// The random numbers generator.
		Random rng = new Random(Model.seed);
		// Fills the performance map with knowledgefields and random performance 
		// in [0,1].
		for (KnowledgeField kf : knowledgeFields.keySet()) {
			// Random performance needs to be computed for novel knowledge fields only.
			// We don't want that one and the same kf yields a different performance at each time step.
			// The method updateField in this file enters the conventional value -0.1 whenever a novel kf is added to common knowledge.
			// Only when performance is equal to -0.1 a new random value is computed.
			if (knowledgeFields.get(kf).performance == -0.1) {  
				int kfi = knowledgeFields.get(kf).KFIs;  
				// stores the number of instances of kf
				knowledgeFields.put(kf, new CommonKnowledgeData 
						(kfi, rng.nextDouble()));
			}
		}
	}
	
	/**
	 * Computes the Betweeness Centrality. Employs the library 
	 * http://jung.sourceforge.net. Based on: Ulrik Brandes, "A Faster Algorithm 
	 * for Betweeness Centrality". Journal of Mathematical Sociology, 
	 * 25(2): 163-177, 2001.
	 * 
	 * Computes betweenness centrality for each vertex and edge in the graph. 
	 * The result is that each vertex and edge has a UserData element of type 
	 * MutableDouble whose key is 'centrality.BetweennessCentrality'. 
	 * Note: Many social network researchers like to normalize the betweenness 
	 * values by dividing the values by (n-1)(n-2)/2. The values given here are 
	 * unnormalized.
	 */
	private void computeBetweenness() {
		// Defines the graph.
		UndirectedSparseGraph<KnowledgeField, String> graph = 
				new UndirectedSparseGraph<KnowledgeField, String>();
		
		// Generate the graph and return how many nodes are connected to some 
		// other node
		int numCN = this.setupGraphAndGetNumCN(graph);

		// Betweenness centrality is evaluated.
		BetweennessCentrality<KnowledgeField, String> ranker = 
				new BetweennessCentrality<KnowledgeField, String>(graph);
		ranker.setRemoveRankScoresOnFinalize(false);
		ranker.evaluate();
		
		// Betweenness centrality is retrieved and normalized to the [0,1] 
		// interval.
		for (KnowledgeField kf : knowledgeFields.keySet()) {
			// knowledge field instances in CommonKnowledge
			int kfis = knowledgeFields.get(kf).KFIs;
		
			// not yet normalized
			double bc = ranker.getVertexRankScore(kf);  
			if (bc > 0 && numCN > 2)
				// normalized betweenness centrality
				bc = bc * 2 / ((numCN - 1) * (numCN - 2));  
			
			knowledgeFields.put(kf, new CommonKnowledgeData(kfis, bc));
//System.out.println(this.getPerformance(kf));
		}
	}
	
	/**
	 * Computes the Degree Centrality. Employs the library 
	 * http://jung.sourceforge.net
	 * http://jung.sourceforge.net/doc/api/edu/uci/ics/jung/algorithms/scoring/DegreeScorer.html 
	 */
	private void computeDegreeCentrality() {
		//System.out.println("computeDegreeCentrality");
		// Defines the graph.
		UndirectedSparseGraph<KnowledgeField, String> graph = 
						new UndirectedSparseGraph<KnowledgeField, String>();
				
		// Generate the graph and return how many nodes are connected to some 
		// other node
		int numCN = this.setupGraphAndGetNumCN(graph);

		// Degree centrality is evaluated.
		DegreeScorer<KnowledgeField> ranker = 
						new DegreeScorer<KnowledgeField>(graph);
		//ranker.setRemoveRankScoresOnFinalize(false);
		//ranker.evaluate();
				
		// Degree centrality is retrieved and normalized to the [0,1] interval.
		for (KnowledgeField kf : knowledgeFields.keySet()) {
					// knowledge field instances in CommonKnowledge
			int kfis = knowledgeFields.get(kf).KFIs;
				
			// not yet normalized
			double dc = ranker.getVertexScore(kf);  
					
			if (dc > 0 && numCN > 2)
				// normalized degree centrality
				dc = dc / (numCN - 1);  
					
			knowledgeFields.put(kf, new CommonKnowledgeData(kfis, dc));
//System.out.println(this.getPerformance(kf));
		}
	}
	
	/**
	 * Set up graph, adding nodes and edges. Return size of connected nodes.
	 * Called when computing betweenness centrality.
	 * 
	 * @param graph UndirectedSparseGraph
	 * @return Return number of connected nodes.
	 */
	private int setupGraphAndGetNumCN(UndirectedSparseGraph graph) {
		
		// Fills the graph with the nodes, but not yet the edges.
		for (KnowledgeField kf : knowledgeFields.keySet()) {
			graph.addVertex(kf);
		}
		
		// Transfers knowledge fields into a list in order to scan them with two 
		// indices.
		ArrayList<KnowledgeField> KFsList = new ArrayList<KnowledgeField>();
		for (KnowledgeField kf : knowledgeFields.keySet()) {
			KFsList.add(kf);
		}
		
		// Identifies edges and puts them in the graph. 
		// Takes note of connected nodes.
		
		// Fill graph with edges.
		// non-isolated nodes
		ArrayList<Integer> connectedNodes = new ArrayList<Integer>();  
		for (int i=0; i<KFsList.size(); i++)
			for (int j=0; j<i; j++) {
				KnowledgeField kfi = KFsList.get(i);
				KnowledgeField kfj = KFsList.get(j);
				if (kfi.intersects(kfj)) {
					String nome_edge = 
							Integer.toString(i)+","+Integer.toString(j);
					graph.addEdge(nome_edge, kfi, kfj);
					if (!connectedNodes.contains(i))
						connectedNodes.add(i);
					if (!connectedNodes.contains(j))
						connectedNodes.add(j);
				}
		}
		
		return connectedNodes.size();
	}
}