
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;


/*!
 * Contains knowledge fields and their depths.
 * 
 *  \author Sirio Capizzi
 *  \author Guido Fioretti
 *  \author Ruggero Rossi
 *  \copyright GNU Public License.
 */
public class KnowledgeBase {
	
	protected double maxExpertise; ///< The maximum expertise allowed.
	
	/**
	 * 
	 * @param maxE
	 */
	public KnowledgeBase(double maxE) {
		maxExpertise = maxE;
	}
	
	/**
	 *  A firm's knowledge is made of items composed by a knowledge field, 
	 *  its current expertise, and the maximum expertise it ever attained. 
	 *  It is organized as a hashmap with Key: KnowledgeField, and 
	 *  Value: ExpertiseData, where ExpertiseData is composed of current 
	 *  expertise and the maximum expertise ever attained.
	 */
	private HashMap<KnowledgeField,ExpertiseData> knowledge = 
			new HashMap<KnowledgeField,ExpertiseData>();
	
	/**
	 *  Checks whether a specific knowledge field is there.
	 *  It must check BEIs because it is eventually invoked before KFs' IDs 
	 *  have been fixed. (by Firm, at knowledge initialization)
	 *
	 * @param kf
	 * @return
	 */
	public boolean isThereField(KnowledgeField kf) {
		boolean isThere = false;
		// Scans its knowledge fields.
		for (KnowledgeField j : knowledge.keySet()) {
			boolean sameBEIs = true;
			// Scans all the pairs of BEIs of knowledge fields j and kf.
			for (int i=0; i<Model.numBusinessElements; i++)
				if (j.getInstance(i).getInstanceOfAspect() != kf.getInstance(i).getInstanceOfAspect())  
					// BEIs' IDs
					sameBEIs = false;
			if (sameBEIs == true)
				isThere = true;
		}	
		return isThere;
	}
	
	/**
	 * Checks how many replicas of a specific BEI are in this knowledge base.
	 *
	 * @param BEI_ID
	 * @return
	 */
	public int howManyOfThisBEI(long BEI_ID) {
		int howMany = 0;
		// Scans its knowledge fields.
		for (KnowledgeField j : knowledge.keySet()) {
			// Scans the business element instances within each knowledge field.
			for(int i=0; i<Model.numBusinessElements; i++) {
				BusinessElementInstance a_bei = j.getInstance(i);
				if (a_bei.getInstanceOfAspect() == BEI_ID )
					howMany++;
			}			
		}
		return howMany;
	}
	
	/**
	 * Returns the number of fields.
	 *
	 * @return
	 */
	public int getNumFields() {
		return knowledge.size();
	}

	/**
	 * Gets an expertise value. If the field is not there, it returns zero.
	 * 
	 * @param kf
	 * @return
	 */
	public double getExpertise(KnowledgeField kf) {
		Double d = knowledge.get(kf).expertise;
		return (d!=null)?d:0f;
	}
	
	/**
	 * Get the lowest, the average and the highest expertise.
	 * Get the knowledge field with lowest expertise and the knowledge field 
	 * with highest expertise.
	 *
	 * @return
	 */
	public double getLowestExpertise() {
		double lowestExpertise = maxExpertise;
		ArrayList<Map.Entry<KnowledgeField,ExpertiseData>> myK = 
				new ArrayList<Map.Entry<KnowledgeField,ExpertiseData>>(this.getKnowledge());
		for(Map.Entry<KnowledgeField,ExpertiseData> e: myK) {
			if (e.getValue().expertise < lowestExpertise)
			lowestExpertise = e.getValue().expertise;
		}
		return lowestExpertise;
	}
	
	/**
	 * 
	 * @return
	 */
	public KnowledgeField getKFwithLowestExpertise() {
		double lowestExpertise = maxExpertise;
		KnowledgeField lowestEkf = null;
		ArrayList<Map.Entry<KnowledgeField,ExpertiseData>> myK = 
				new ArrayList<Map.Entry<KnowledgeField,ExpertiseData>>(this.getKnowledge());
		for(Map.Entry<KnowledgeField,ExpertiseData> e: myK) {
			if (!e.getKey().entailsUndefinedBEI() && e.getValue().expertise < lowestExpertise) {
				lowestExpertise = e.getValue().expertise;
				lowestEkf = e.getKey();
			}
		}
		return lowestEkf;
	}
	
	/**
	 * 
	 * @return
	 */
	public double getAverageExpertise() {
		double sum = 0.0f;
		double aveExpertise = 0.0f;
		ArrayList<Map.Entry<KnowledgeField,ExpertiseData>> myK = 
				new ArrayList<Map.Entry<KnowledgeField,ExpertiseData>>(this.getKnowledge());
		for(Map.Entry<KnowledgeField,ExpertiseData> e: myK) {
			sum += e.getValue().expertise;
		}
		if (knowledge.size() > 0)
			aveExpertise = sum/(double)knowledge.size();
		return aveExpertise;
	}
	
	/**
	 * 
	 * @return
	 */
	public double getHighestExpertise() {
		double highestExpertise = 0.0f;
		ArrayList<Map.Entry<KnowledgeField,ExpertiseData>> myK = 
				new ArrayList<Map.Entry<KnowledgeField,ExpertiseData>>(this.getKnowledge());
		for(Map.Entry<KnowledgeField,ExpertiseData> e: myK) {
			if (e.getValue().expertise > highestExpertise)
			highestExpertise = e.getValue().expertise;
		}
		return highestExpertise;
	}
	
	/**
	 * 
	 * @return
	 */
	public KnowledgeField getKFwithHighestExpertise() {
		double highestExpertise = 0.0f;
		KnowledgeField highestExpertiseKF = null;
		ArrayList<Map.Entry<KnowledgeField,ExpertiseData>> myK = 
				new ArrayList<Map.Entry<KnowledgeField,ExpertiseData>>(this.getKnowledge());
		for(Map.Entry<KnowledgeField,ExpertiseData> e: myK) {
			if (e.getValue().expertise > highestExpertise) {
				highestExpertise = e.getValue().expertise;
				highestExpertiseKF = e.getKey();
			}
		}
		return highestExpertiseKF;
	}

	/**
	 * Returns the whole knowledge base.
	 *
	 * @return
	 */
	public Set<Map.Entry<KnowledgeField,ExpertiseData>> getKnowledge() {
		return knowledge.entrySet();
	}
	
	/**
	 * Tunes the expertise of a knowledge field by increasing or decreasing it 
	 * by an amount
	 * which is a fraction of the maximum expertise ever attained by that field.
	 *
	 * @param kf
	 * @param fraction
	 */
	public void tuneFieldExpertise(KnowledgeField kf, double fraction) {
		if (kf != null) {
			double newExpertise;
			double newExpertiseAttained;
            // Updates expertise.
			newExpertise = knowledge.get(kf).expertise;  	// initial value of expertise
			newExpertise += fraction*knowledge.get(kf).expertiseAttained;  // its variation
			// Curbs excess values.
			if (newExpertise > maxExpertise)
				newExpertise = maxExpertise;
			else if (newExpertise < 0)
				newExpertise = 0;
            // Updates maximum expertise.
			newExpertiseAttained = knowledge.get(kf).expertiseAttained;
			if (newExpertise > newExpertiseAttained)
				newExpertiseAttained = newExpertise;
			// Uploads the new values.
			knowledge.put(kf, 
					new ExpertiseData(newExpertise,newExpertiseAttained));
		}
        else {
            System.out.println("Problem at tuneFieldExpertise");
        }
	}

	/**
	 * Adds a new knowledge field with its expertise, or resets the expertise 
	 * of an existing one.
	 * The greatest expertise ever attained is reset as well.
	 *
	 * @param kf
	 * @param d
	 */
	public void setField(KnowledgeField kf, double d) {
		double newExpertise = d;
		if (newExpertise > maxExpertise)
			newExpertise = maxExpertise;
		else if (newExpertise < 0)
			newExpertise = 0;
		if (kf != null) {
			knowledge.put(kf, new ExpertiseData(newExpertise, newExpertise));
		}
	}
	
	/**
	 * Removes just one instance of one knowledge field.
	 * 
	 * @param kf
	 */
	public void removeField(KnowledgeField kf) {
		knowledge.remove(kf);
	}
	
	/**
	 * Removes all instances of all knowledge fields in this Knowledge Base.
	 */
	public void removeAllFields() {
		Set<Map.Entry<KnowledgeField,ExpertiseData>> kS = knowledge.entrySet();
		kS.removeAll(kS);
	}
	
	/**
	 * Checks that a Knowledge Base is empty.
	 */
	public boolean isEmpty() {
		return knowledge.isEmpty();
	}

	/**
	 * Returns all knowledge fields.
	 *
	 * @return
	 */
	public Set<KnowledgeField> fields() {
		return knowledge.keySet();
	}
	
	/**
	 * Takes account of the effect of distance on expertise.
	 * The expertise of all fields are multiplied by a distance coefficient.
	 */
	protected void accountForDistance(double dC) {
		Set<Map.Entry<KnowledgeField,ExpertiseData>> kS = knowledge.entrySet();
		for(Map.Entry<KnowledgeField,ExpertiseData> i: kS) {
			setField(i.getKey(), i.getValue().expertise*dC);
			setField(i.getKey(), i.getValue().expertiseAttained*dC);
		}
	}
	
	/**
	/**
	 * Yields the maxExpertise that has received as parameter.
	 *
	 * @return
	 */
	public double getMaxExpertise() {
		return maxExpertise;
	}
}
