
import java.io.*;
import java.util.Random;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.TreeMap;

import org.jfree.chart.ChartPanel;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.JFreeChart;
import org.jfree.ui.ApplicationFrame;
import org.jfree.ui.RefineryUtilities;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.data.category.DefaultCategoryDataset;

 /* per vedere i colori in Pajek: nella finestra di Draw selezionare menu 
  * Options -> Colors -> Edges -> As defined on input file 
  */

/*!
 * Model manages all other agents and objects.
 * 
 *  \author Sirio Capizzi
 *  \author Guido Fioretti
 *  \author Ruggero Rossi
 *  \copyright GNU Public License.
 */
public class Model {

	//PARAMETERS DECLARATION

	// Simulation Parameters.
	public static int stopsAt = 100; ///< the number of simulation steps
	public static boolean changeSeed = false; ///< if true, random seed for the random numbers generators
	public static long seed = 107965; 	///< the seed of the random numbers generators
	public static int saveNetworkStep = 100;  ///< the intervals at which network data are written
	public static int simulation_number = 1; 	///< the number of simulations, always = 1 in single-run 
												///< (this is used in the names of the graphs)
	
	// Model Parameters.
	public int worldLength = 100; 	///< the world is a torus obtained from a square
	public int numClusters = 1;	///< the number of clusters
	public int numClusteredInnovators = 10;	///< the number of clustered innovators
	public int numClusteredImitators = 10;	///< the number of clustered imitators
	public int numIsolatedInnovators = 0; 	///< the number of isolated innovators
	public int numIsolatedImitators = 0; 	///< the number of isolated imitators
	
	// Firms Parameters.
	public int maxFieldsPerFirm = 2;	///< the maximum number of knowledge 
										///< fields that a firm can manage
	public float maxExpertise = 10.0f;	///< the maximum expertise that can be 
										///< attained on a knowledge field
	public static int numBusinessElements = 2;	///< the number of aspects of knowledge:
												///< product, market, technology,
												///< organization, etc.
	public static int newBusinessElements = 1;	///< in how many aspects 
												///< novel instances are created when 
												///< original exploration is made
	public static int originOfPerformance = 1; 	///< 0: Random performance; 
												///< 1: Betweenness centrality;
												///< 2: Degree centrality.
	public float decayRate = 0.1f;	///< the rate of decay of capital and 
									///< knowledge
	public float cognitiveDistanceWindow = 0.5f;///< the central window of 
												///< intermediate cognitive 
												///< distance
	public float maxCognitiveEffort = 1.0f;	///< the maximum cognitive effort 
											///< that can be spent to take 
											///< account of rivals
	
	// Datasets for the line charts to be drawn during a simulation.
	DefaultCategoryDataset datasetRivals = new DefaultCategoryDataset( );
	DefaultCategoryDataset datasetActions = new DefaultCategoryDataset( );
	DefaultCategoryDataset datasetCognitiveDistance = new DefaultCategoryDataset( );
	DefaultCategoryDataset datasetPerformance = new DefaultCategoryDataset( );
	DefaultCategoryDataset datasetKnowledgeFields = new DefaultCategoryDataset( );
	DefaultCategoryDataset datasetExpertise = new DefaultCategoryDataset( );
	DefaultCategoryDataset datasetDeaths = new DefaultCategoryDataset( );

	// Initialization Parameters.
	double minInitialSize = 1.0;  ///< the minimum initial size of firms
	double maxInitialSize = 1000.0; ///< the maximum initial size of firms
	double sizeDistrSlope = -0.5;	///< the slope of the power law of firm size: 
									///< at initialization, and when firms are substituted
	static int varietyAtInitialization = 2;	///< the number of BEIs 
												///< initially created for each 
												///< business element.
	
	//VARIABLES DECLARATION

	// Simulation variables.
	public static int step; ///< the counter of simulation steps
	public static int run;  ///< the counter of simulation runs
	public int newBEIsIDs = maxFieldsPerFirm + numBusinessElements* 
			(numIsolatedInnovators+numIsolatedImitators + numClusteredInnovators + numClusteredImitators); 
	///< the starting point for the IDs of new BEIs created during the 
	///< simulation, read and updated each time a new BEI is created
	///< the number of business elements multiplied by the number of firms

	// The number of deaths (and substitutions).
	public int clusteredImitatorDeaths = 0;	///< the number of times an Imitator located in a cluster dies
	public int isolatedImitatorDeaths = 0;	///< the number of times an isolated Imitator dies
	public int clusteredInnovatorDeaths = 0;///< the number of times an Innovator located in a cluster dies
	public int isolatedInnovatorDeaths = 0; ///< the number of times an isolated Innovator dies
	
	//DECLARATION OF MISCELLANEOUS OBJECTS
	
	// The list of firms, the list of all possible BEIs, and common knowledge.
	public static ArrayList<Firm> firmList; ///< the list of firms
	public static ArrayList<BusinessElementInstance>[] instances; 
		///< a vector of lists of BEIs, one for each business element
	public CommonKnowledge commonKnowledge; ///< stores the number of instances 
											///< of each knowledge field and 
											///< computes the betweeness 
											///< centrality of the graph of 
											///< knowledge fields, which 
											///< measures performance

	// Action and HistoricAction networks.
	ArrayList <FirmAction>  firms_actions;	///< the actions made by firms in 
											///< this step
	ArrayList <FirmHistoricAction>  firms_historic_actions;	///< the actions 
											///< made by firms since last graph 
											///< savings
	HashMap <Integer, HistoricFirm> firms_historic; ///< the firms who did 
											///< actions since last graph 
											///< savings
	/**
	 * 
	 */
	class HistoricFirm {
		int clusterNumber;	///< cluster ID
		java.awt.Color color;	///< a firm's color tells whether it's an 
								///< Innovator or an Imitator
		int xPos;
		int yPos;
	}

	// To collect and write the outcomes of the model.
	public BufferedWriter stats;  ///< time series
	public FileWriter BEIs;	///< business element instances at each simulation step
	public FileWriter KFs;  ///< knowledge fields at each simulation step

	//BUILDS THE NETWORKS
	 
	/**
	 * Knowledge Network
	 * Nodes are knowledge fields, that are linked if they share an instance of 
	 * a business element.
	 * All nodes are black. Their labels show knowledge fields' IDs.
	 * Edges are colored depending on which business element any two knowledge 
	 * fields have in common:
	 * - BE1: Blue;
	 * - BE2: Yellow;
	 * - BE3: Red;
	 * - BE4: Green;
	 * - BE5: Purple.
	 *
	 * @return
	 */
	public LayeredMatrix GenerateKnowledgeNetwork() {
		int n_fields = commonKnowledge.getKnowledgeFields().size();
		LayeredMatrix lm = new LayeredMatrix(n_fields, numBusinessElements);

		int m = 0;
		for (KnowledgeField j : commonKnowledge.getKnowledgeFields().keySet()) {
			int n = 0;
			for (KnowledgeField k : commonKnowledge.getKnowledgeFields().keySet()) {
				lm.m[m][n] = 0;
				if (j != k)
					for (int i = 0; i < numBusinessElements; i++) {
						if (j.getInstance(i) == k.getInstance(i)) {
							lm.m[m][n] |= (1 << i); // settiamo il bit
						}
					}
				n++;
			}
			m++;
		}

		return lm;
	}
	
	/**
	 *  salva il grafo delle intersezioni tra i BEI della CommonKnowledge
	 * 
	 * @param network
	 * @param filename
	 * @return Returns 0 if ok
	 */
	public boolean WriteKnowledgeGraph(LayeredMatrix network, String filename) {
		FileOutputStream fos;
		PrintStream ps;
		File f = new File(filename);
		boolean error = false;

		// nota: \n è LF, prima bisogna mettere CR per farlo leggere da Pajek.
		// "bufferedInputStream.ReadLine()" non ha problemi aggiungendo CR

		try {
			fos = new FileOutputStream(f);
		} catch (Exception e) {
			System.out.println("\nGraphFile write error on opening: " + e);
			return false;
		}

		ps = new PrintStream(fos);

		// /////////////////// qua in mezzo scriviamo le cose ///////////////
		Object obj[]= commonKnowledge.getKnowledgeFields().keySet().toArray();

		int total_nodes = network.m.length;
		ps.print("*Vertices      " + total_nodes + "\r\n");

		for (int i = 0; i < total_nodes; i++) {
			int ord = i + 1;
			long num=((KnowledgeField)obj[i]).getID();
			ps.print("   " + ord + " \"" + num + " \""	// the labels of the 
														// nodes
					+ " ic Black" + " bc Black\r\n");  	// the colors of the 
														// nodes (ic), and their 
														// rims (bc)
		}

		ps.print("*Edges\r\n");
		for (int i = 0; i < total_nodes; i++) {
			int num = i + 1;
			for (int j = i + 1; j < total_nodes; j++) {
				int num2 = j + 1;
				for (int layer = 0; layer < network.n_layers; layer++) {
					if ((network.m[i][j] & (1 << layer)) > 0) {
						ps.print("       " + num + "        " + num2 + " 1 c ");
						switch (layer) {
						case 0:
							ps.print("Blue");
							break;
						case 1:
							ps.print("Yellow");
							break;
						case 2:
							ps.print("Red");
							break;
						case 3:
							ps.print("Green");
							break;
						case 4:
							ps.print("Purple");
							break;
						}
						ps.print("\r\n");
					}
				}
			}
		}
		// //////////////////////////////////////////////////////////////////

		ps.close();

		try {
			if (fos != null)
				fos.close();
		} catch (Exception e) {
			System.out.println("\nGraphFile write close error: " + e);
			return false;
		}

		if (error)
			return false;

		return true;
	}  // End of Knowledge Network


	/**
	 * The network of actions: Action
	 * The network of past actions: HistoricActions
	 * Nodes represent firms. Their labels tell their IDs. The labels of 
	 * clustered firms also show their cluster's ID preceded by a 'C'.
	 * Actions are illustrated by nodes' colors and edges' colors.
	 * Nodes representing Innovators are black if they undertake no action. 
	 * Their rim is also black.
	 * - If Innovators undertake Experiential Exploration: Red node.
	 * - If Innovators undertake Experiential Exploitation: Blue rim.
	 * Nodes representing Imitators are always gray, with gray rim.
	 * Vicarious learning is represented by edge color for both Innovators and 
	 * Imitators:
	 * - Vicarious Exploration: Red edge.
	 * - Vicarious Exploitation: Blue edge.
	 *
	 * @param filename
	 * @return
	 */
	public boolean WriteActionGraph(String filename) {
		FileOutputStream fos;
		PrintStream ps;
		File f = new File(filename);
		boolean error = false;

		// nota: \n è LF, prima bisogna mettere CR per farlo leggere da Pajek.
		// "bufferedInputStream.ReadLine()" non ha problemi aggiungendo CR

		try {
			fos = new FileOutputStream(f);
		} catch (Exception e) {
			System.out.println("\nGraphFile write error on opening: " + e);
			return false;
		}

		ps = new PrintStream(fos);

		/////////////////////// writes the graph ///////////////
		// Experiential learning: firms act on themselves.
		boolean exploration_self[]=new boolean[firmList.size()];
		boolean exploitation_self[]=new boolean[firmList.size()];

		for(FirmAction action: firms_actions)
		{
			if(action.actor==action.inspiration) // azione con s� stessa
			{
				int code=0;
				for (int i = 0; i < firmList.size(); i++) {
					Firm firm=((Firm)firmList.get(i));
					if(firm==action.actor)
					{
						code=i;
						break;
					}				
				}
				
				if(action.action==FirmAction.action_exploration)
					exploration_self[code]=true;
				else if(action.action==FirmAction.action_exploitation)
					exploitation_self[code]=true;
			}		
		}
		
		// Actions. The nodes.
		int total_nodes = firmList.size();
		ps.print("*Vertices      " + total_nodes + "\r\n");

		for (int i = 0; i < total_nodes; i++) {
			int ord = i + 1;  // numbers the nodes
			Firm a_firm=((Firm)firmList.get(i));
			int num=a_firm.getIdentity();
			int itsCluster = a_firm.cluster;

			// Writes nodes' labels.
			ps.print("   " + ord + " \"");  // the node's number
			// If it's an isolated firm, it writes its ID.
			// If it's a clustered firm, it also writes a C followed by 
			// cluster's ID.
			if (itsCluster == 0)
				ps.print(""+ num + " \" " +((float)a_firm.xPos)/worldLength+
						" "+((float)a_firm.yPos)/worldLength+" 0.5000 ");
			else
				ps.print(""+ num + " C" + itsCluster + 
						" \" " +((float)a_firm.xPos)/worldLength+
						" "+((float)a_firm.yPos)/worldLength+" 0.5000 ");
			
			// Writes nodes' colors.
			if (a_firm.getColor() == java.awt.Color.gray) {  
				// if it's an Imitator
				ps.print(" ic Gray");  // the node is always gray
				ps.print(" bc Gray");  // its rim is always gray
			}
			else {  // if it's an Innovator				
				if((!exploration_self[i]))
					ps.print(" ic Black");	// default node color
				else
					ps.print(" ic Red");	// if experiential exploration, 
											// red node
				if(exploitation_self[i])	// if experiential exploitation, 
											// blue rim
					ps.print(" bc Blue");
				else
					ps.print(" bc Black");  // default rim color
			}
			
			ps.print("\r\n");
		}

		ps.print("*Arcs\r\n");
		
		for(FirmAction action: firms_actions)
		{
			
			int code1=0,code2=0;
			for (int i = 0; i < total_nodes; i++) {
				Firm firm=((Firm)firmList.get(i));
				if(firm==action.actor)
				{
					code1=i+1;
					break;
				}				
			}
			for (int i = 0; i < total_nodes; i++) {
				Firm firm=((Firm)firmList.get(i));
				if(firm==action.inspiration)
				{
					code2=i+1;
					break;
				}				
			}
			
			ps.print("       " + code1 + "        " + code2 + " 1 c ");
			if(action.action==FirmAction.action_exploration)
				ps.print("Red");  // if vicarious exploration, red edge
			else if(action.action==FirmAction.action_exploitation)
				ps.print("Blue");  // if vicarious exploitation, blue edge
			ps.print("\r\n");
		}
		// //////////////////////////////////////////////////////////////////

		ps.close();

		try {
			if (fos != null)
				fos.close();
		} catch (Exception e) {
			System.out.println("\nGraphFile write close error: " + e);
			return false;
		}

		if (error)
			return false;

		return true;
	}  // end of Action
	
	/**
	 *  HistoricAction begins.
	 *  Salva il grafo delle intersezioni tra i BEI delle aziende.
	 *   
	 * @param filename
	 * @return Returns 0 if ok
	 */
	public boolean WriteAllActionsGraph(String filename) {
		FileOutputStream fos;
		PrintStream ps;
		File f = new File(filename);
		boolean error = false;

		// nota: \n è LF, prima bisogna mettere CR per farlo leggere da Pajek.
		// "bufferedInputStream.ReadLine()" non ha problemi aggiungendo CR

		try {
			fos = new FileOutputStream(f);
		} catch (Exception e) {
			System.out.println("\nGraphFile write error on opening: " + e);
			return false;
		}

		ps = new PrintStream(fos);

		//////////////////// writes the graph ///////////////
		int total_nodes = firms_historic.size();
		int total_firms[]=new int[total_nodes];
		
		ps.print("*Vertices      " + total_nodes + "\r\n");
		int i=0;
		
		// leggiamo il codice-nome di ogni impresa
		for (Integer j : firms_historic.keySet())	
		{
			total_firms[i]=j.intValue();
			i++;			
		}

		// prima guardiamo quali sono le imprese che fanno azioni con se stessa
		boolean exploration_self[]=new boolean[total_nodes];
		boolean exploitation_self[]=new boolean[total_nodes];
		for(FirmHistoricAction action: firms_historic_actions)
		{
			if(action.actor_id==action.inspiration_id) // azione con se stessa
			{
				int code=0;
				for (int n = 0; n < total_nodes; n++) {
					if(total_firms[n]==action.actor_id)
					{
						code=n;
						break;
					}				
				}
				
				if(action.action==FirmAction.action_exploration)
					exploration_self[code]=true;
				else if(action.action==FirmAction.action_exploitation)
					exploitation_self[code]=true;
			}
		}
		
		// scriviamo le imprese
		for (i=0; i<total_nodes; i++)
		{		
			// Nodes labels.
			int n=i+1;  // numbers the nodes
			ps.print("   " + n + " \"");  // writes the node's number
			HistoricFirm theFirm = firms_historic.get(total_firms[i]);  
				// total_firms[i] is a firm's ID
			
			// All labels display firms IDs. The labels of clustered firms 
			// display also cluster ID.
			if (theFirm.clusterNumber > 0)	// if it is clustered
				ps.print(""+ total_firms[i] + " C" + theFirm.clusterNumber + 
						" \" " +((float)theFirm.xPos)/worldLength+
						" "+((float)theFirm.yPos)/worldLength+" 0.5000 ");
			else
				ps.print(""+ total_firms[i] + 
						" \" " +((float)theFirm.xPos)/worldLength+
						" "+((float)theFirm.yPos)/worldLength+" 0.5000 ");

			ps.print(""+ total_firms[i] + " \" " );
			
			// HistoricAction. Nodes colors.
			if (theFirm.color == java.awt.Color.gray) {  // if it's an Imitator
				ps.print(" ic Gray");  // the node is always gray
				ps.print(" bc Gray");  // as well as its rim
			}
			else {  // if it's an Innovator
				if((!exploration_self[i]))
					ps.print(" ic Black");	// default node color if no action 
											// is undertaken
				else
					ps.print(" ic Red");	// if experiential exploration, 
											// red node
				if(exploitation_self[i])	// if experiential exploitation blue 
											// rim
					ps.print(" bc Blue");
				else
					ps.print(" bc Black");  // default rim color if no action is 
											// undertaken			
			}
			
			ps.print("\r\n");
		}
		
		ps.print("*Arcs\r\n");
		
		for(FirmHistoricAction action: firms_historic_actions)
		{
			
			int code1=0,code2=0;
			for (int n = 0; n < total_nodes; n++) {
				if(total_firms[n]==action.actor_id)
				{
					code1=n+1;
					break;
				}				
			}
			for (int n = 0; n < total_nodes; n++) {
				if(total_firms[n]==action.inspiration_id)
				{
					code2=n+1;
					break;
				}				
			}
			
			// HistoricActions. Edges colors.
			ps.print("       " + code1 + "        " + code2 + " 1 c ");
			if(action.action==FirmAction.action_exploration)
				ps.print("Red");
			else if(action.action==FirmAction.action_exploitation)
				ps.print("Blue");
			ps.print("\r\n");
		}
		///////////////////////////////////////////////////////////////////

		ps.close();

		try {
			if (fos != null)
				fos.close();
		} catch (Exception e) {
			System.out.println("\nGraphFile write close error: " + e);
			return false;
		}

		if (error)
			return false;

		return true;
	} // end of action networks
	
	/**
	 *  This method is called by firms in order to store any action they 
	 *  undertake in the matrix of past actions.
	 * 
	 * @param actor
	 * @param inspiration
	 * @param action_type
	 */
	public void InsertHistoricAction(Firm actor, Firm inspiration, 
			int action_type)
	{
		FirmHistoricAction action=new FirmHistoricAction(this, actor, 
				inspiration, action_type);
		firms_historic_actions.add(action);
		
		HistoricFirm theFirm=new HistoricFirm();
		theFirm.clusterNumber = actor.cluster;
		theFirm.color = actor.getColor();
		theFirm.xPos=actor.xPos;
		theFirm.yPos=actor.yPos;
		firms_historic.put(new Integer (actor.getIdentity()), theFirm);
		
		HistoricFirm theFirm2=new HistoricFirm();
		theFirm2.clusterNumber = inspiration.cluster;
		theFirm2.color = inspiration.getColor();
		theFirm2.xPos=inspiration.xPos;
		theFirm2.yPos=inspiration.yPos;
		firms_historic.put(new Integer (inspiration.getIdentity()), theFirm2);
	}
	
	/**
	 * The Similarity Network.
	 * The Similarity Network depicts a similarity index between firms' 
	 * knowledge bases at a particular simulation step.
	 * Nodes represent firms. All nodes are black.
	 * Labels show firms' IDs. The labels of clustered firms also show their 
	 * cluster's number preceded by a 'C'.
	 * The similarity index between two firms' knowledge bases is illustrated 
	 * by edge color.
	 * - If 0.0 <= Similarity Index <  0.2 --> No edge
	 * - If 0.2 <= Similarity Index <  0.4 --> Yellow edge
	 * - If 0.4 <= Similarity Index <  0.6 --> Orange edge
	 * - If 0.6 <= Similarity Index <  0.8 --> Red edge
	 * - If 0.8 <= Similarity Index <= 1.0 --> Violet edge
	 *
	 * @return Returns a matrix representing the network between all firms.
	 * The matrix is an adiacency matrix: the element of a (row,column) is the
	 * weight of the link between the rowth-firm and the columnth-firm
	 * that weight is calculating summing the intersection values between the 
	 * two firms. Hence it is a symmetric matrix. The diagonal is posed = 0 (no 
	 * link with themselves).
	 */
	public double[][] GenerateSimilarityNetwork() {
		int n_firms = firmList.size();
		double links[][] = new double[n_firms][n_firms];

		for (int m = 0; m < n_firms; m++)
			for (int n = 0; n < n_firms; n++)
				if (m != n)
					links[m][n] = ((Firm) firmList.get(m)).getValueSI(n);
				else
					links[m][n] = 0.0f;

		return links;
	}

	/**
	 * Writes the similarity network. Returns 0 if it worked.
	 * 
	 * @param network
	 * @param filename
	 * @return
	 */
	public boolean WriteSimilarityGraph(double[][] network, String filename) {
		// Opens input/output files
		FileOutputStream fos;
		PrintStream ps;
		File f = new File(filename);
		boolean error = false;
		try {
			fos = new FileOutputStream(f);
		} catch (Exception e) {
			System.out.println("\nGraphFile write error on opening: " + e);
			return false;
		}
		ps = new PrintStream(fos);

		// /////////////////// qua in mezzo scriviamo le cose ///////////////

		int total_nodes = network.length;
		ps.print("*Vertices      " + total_nodes + "\r\n");

		for (int i = 0; i < total_nodes; i++) {
			int ord=i+1;  // numerates nodes
				
			// Gets a firm's identity
			Firm theFirm=(Firm)firmList.get(i);
			int num=theFirm.getIdentity();
				
			// Assigns identities to nodes.
			ps.print("   " + ord + " \"" + num );
				
			// Adds cluster number to node labels. Assigns colors to the nodes 
			// and their rims.
			if(theFirm.cluster==0)
				ps.print(" \" " +((float)theFirm.xPos)/worldLength+
						" "+((float)theFirm.yPos)/worldLength+" 0.5000 "
						+ " ic Black" + " bc Black\r\n");
			else
				ps.print(" C" + theFirm.cluster + 
						" \" " +((float)theFirm.xPos)/worldLength+
						" "+((float)theFirm.yPos)/worldLength+" 0.5000 " 
						+ " ic Black" + " bc Black\r\n");
		}

		// Draws edges.
		ps.print("*Edges\r\n");
		for (int i = 0; i < total_nodes; i++) {
			int num = i + 1;
			for (int j = i + 1; j < total_nodes; j++) {
				int num2 = j + 1;
				String linea="       " + num + "        " + num2 + 
						" " 	+ network[i][j];
				if ((network[i][j] >= 0.2)&&(network[i][j] < 0.4))
					ps.print(linea + " c Yellow\r\n");
				else if ((network[i][j] >= 0.4)&&(network[i][j] < 0.6))
					ps.print(linea + " c Orange\r\n");
				else if ((network[i][j] >= 0.6)&&(network[i][j] < 0.8))
					ps.print(linea + " c Red\r\n");
				else if (network[i][j] >= 0.8)
					ps.print(linea + " c Purple\r\n");
			}
		}

		// Close input/output files.
		ps.close();
		try {
			if (fos != null)
				fos.close();
		} catch (Exception e) {
			System.out.println("\nGraphFile write close error: " + e);
			return false;
		}
		if (error)
			return false;
		return true;
	}  // end similarity graph
		
	/**
	 * A series of methods to write information on:
	 * - Action Network;
	 * - HistoricAction Network;
	 * - Knowledge Network;
	 * - Similarity Network.
	 * 
	 * @return
	 */
	public boolean SaveAllActionsGraph() {
		String filename = "output/HistActions_" + run + "_" + (step)
				+ ".net";
		return WriteAllActionsGraph(filename);
	}
	
	/**
	 * 
	 * @return
	 */
	public boolean SaveAllActionsGraphEnd() {
		String filename = "output/HistActions_" + run + "_" + (step)
				+ "_End.net";
		return WriteAllActionsGraph(filename);
	}
	
	/**
	 * 
	 * @return
	 */
	public boolean SaveActionGraph() {
		String filename = "output/Action_" + run + "_" + (step)
				+ ".net";
		return WriteActionGraph(filename);
	}
	
	/**
	 * 
	 * @return
	 */
	public boolean SaveActionGraphEnd() {
		String filename = "output/Action_" + run + "_" + (step)
				+ "_End.net";
		return WriteActionGraph(filename);
	}
	
	/**
	 * 
	 * @return
	 */
	public boolean SaveKnowledgeGraph() {
		LayeredMatrix network = GenerateKnowledgeNetwork();
		String filename = "output/Knowledge_" + run + "_" + (step)
				+ ".net";
		return WriteKnowledgeGraph(network, filename);
	}

	/**
	 * 
	 * @return
	 */
	public boolean SaveKnowledgeGraphEnd() {
		LayeredMatrix network = GenerateKnowledgeNetwork();
		String filename = "output/Knowledge_" + run + "_" + (step)
				+ "_End.net";
		return WriteKnowledgeGraph(network, filename);
	}

	/**
	 * 
	 * @return
	 */
	public boolean SaveSimilarityGraph() {
		double network[][] = GenerateSimilarityNetwork();
		String filename = "output/Simil_" + run + "_" + (step) + ".net";
		return WriteSimilarityGraph(network, filename);
	}

	/**
	 * 
	 * @return
	 */
	public boolean SaveSimilarityGraphEnd() {
		double network[][] = GenerateSimilarityNetwork();
		String filename = "output/Simil_" + run + "_" + (step)
				+ "_End.net";
		return WriteSimilarityGraph(network, filename);
	}
	
	// END OF NETWORKS
	
	
	
	
	
	
	/**
	 * Here, the objects populating the model are created. These include:
	 * - Common knowledge;
	 * - The output files;
	 * - Firms;
	 * - The random numbers generator.
	 */
	public void buildObjects() {

		int xPos, yPos; // the positions where firms are placed
		double initialSize; // the initial size of firms
		int i; // spans firms
		int j; // spans clusters

		// Common knowledge.
		commonKnowledge = new CommonKnowledge();	// the common knowledge of all firms
		
		// Output file: The time series produced by the model.
		try {
			stats = new BufferedWriter(new FileWriter("output/stats.dat", true));
		} catch (IOException e) {
			System.out.println("\n statsInfo file write error on opening: " + e);
		}
		// Output file: The data on business element instances.
		try {
			BEIs = new FileWriter("output/BEIs.dat", true);
		} catch (IOException e) {
			System.out.println("\n BEI Info file write error on opening: " + e);
		}
		// Output file: The data on knowledge fields.
		try {
			KFs = new FileWriter("output/KFs.dat", true);
		} catch (IOException e) {
			System.out.println("\n KF Info file write error on opening: " + e);
		}

		// The random numbers generator. It is used to draw: 
		// (1) The position of clusters and isolated firms from a uniform distribution; 
		// (2) The initial size of firms according to a power law at random.
		if (changeSeed == true) {
			Random rsg = new Random();  // the random seed generator
			seed = rsg.nextInt();
		}
		Random rng = new Random(seed);  // the random numbers generator

		// The list of all firms.
		firmList = new ArrayList<Firm>();

		// The vectors with the position of clustered and isolated firms.
		// Clustered firms overlap.
		int xx[] = new int[numIsolatedInnovators+numIsolatedImitators+numClusters];
		int yy[] = new int[numIsolatedInnovators+numIsolatedImitators+numClusters];
		
		// Creates isolated innovators. Conventionally, they belong to cluster 0.
		for (i=0; i<numIsolatedInnovators; i++) {
			xx[i] = rng.nextInt(worldLength);
			yy[i] = rng.nextInt(worldLength);
			initialSize = (double) powerLaw(minInitialSize, maxInitialSize, sizeDistrSlope);
			Innovator anInnovator = new Innovator(this,	i, initialSize, xx[i], yy[i], 0);
			firmList.add(anInnovator);			
		}
		
		// Creates isolated imitators. Conventionally, they belong to cluster 0.
		for (i=numIsolatedInnovators; i<(numIsolatedInnovators+numIsolatedImitators); i++) {
			xx[i] = rng.nextInt(worldLength);
			yy[i] = rng.nextInt(worldLength);
			initialSize = (double) powerLaw(minInitialSize, maxInitialSize, sizeDistrSlope);
			Imitator anImitator = new Imitator(this, i, initialSize, xx[i], yy[i], 0);
			firmList.add(anImitator);
		}
		
		// Finds cluster positions.
		for (j=0; j<numClusters; j++) {
			xx[numIsolatedInnovators+numIsolatedImitators+j] = rng.nextInt(worldLength);
			yy[numIsolatedInnovators+numIsolatedImitators+j] = rng.nextInt(worldLength);			
		}
		
		// Creates clustered innovators.
		i = numIsolatedInnovators+numIsolatedImitators;  // i is the clustered innovators' IDs
		// Clusters are visited several times, dropping one clustered innovator each time.
		// One additional run is necessary if the division does not yield an integer.
		for (int k=0; k<((int)numClusteredInnovators/numClusters + 1); k++) {
			for (j=0; j<numClusters; j++) {  // (j+1) is the clusters' IDs
				if (i < numIsolatedInnovators+numIsolatedImitators+numClusteredInnovators) {
					initialSize = (double) powerLaw(minInitialSize, maxInitialSize, sizeDistrSlope);
					Innovator anInnovator = new Innovator(this,	i, initialSize, 
							xx[numIsolatedInnovators+numIsolatedImitators+j], 
							yy[numIsolatedInnovators+numIsolatedImitators+j], j+1);
					firmList.add(anInnovator);
					i++;
				}
			}
		}
		
		// Creates clustered imitators.
		i = numIsolatedInnovators+numIsolatedImitators+numClusteredInnovators;  // i is the clustered imitators' IDs
		// Clusters are visited several times, dropping one clustered innovator each time.
		// One additional run is necessary if the division does not yield an integer.
		for (int k=0; k<((int)numClusteredImitators/numClusters + 1); k++) {
			for (j=0; j<numClusters; j++) {  // (j+1) is the clusters' IDs
				if (i < numIsolatedInnovators+numIsolatedImitators+numClusteredInnovators
						+numClusteredImitators) {
					initialSize = (double) powerLaw(minInitialSize, maxInitialSize, sizeDistrSlope);
					Imitator anImitator = new Imitator(this, i, initialSize, 
							xx[numIsolatedInnovators+numIsolatedImitators+j], 
							yy[numIsolatedInnovators+numIsolatedImitators+j], j+1);
					firmList.add(anImitator);
					i++;
				}
			}
		}
		


		
		
		/**
		 * Initialilzations of various objects:
		 * - Business element instances;
		 * - Knowledge;
		 * - Betweenness centrality;
		 * - A matrix recording past actions.
		 */
		// Initializes business element instances and assigns them to business 
		// elements.
		instances = new ArrayList[numBusinessElements];
		for (i = 0; i < numBusinessElements; i++) {
			instances[i] = new ArrayList<BusinessElementInstance>();
			// The number of different BEIs from which initial BEIs are drawn at initialization.
			// It must be smaller than the first name of the new BEIs that will 
			// be eventually created afterwards (newBEIsIDs).
			for (j = 0; j < varietyAtInitialization; j++) {
				instances[i].add(new BusinessElementInstance(i, false));
			}
		}
System.out.println("Sono stati generati i seguenti BEI:");
for (i=0; i<numBusinessElements; i++) {
	for (j = 0; j < varietyAtInitialization; j++) {
	System.out.println("BE = " + i + " ID = " + instances[i].get(j).getInstanceOfAspect());
}
}
System.out.println("Fine della generazione dei BEI");

		// Initialization of knowledge.
		for (i = 0; i < firmList.size(); i++) {
			Firm aFirm = (Firm) firmList.get(i);
			aFirm.initializeKnowledge(rng);
		}
		
		// Initialization of performance.
		commonKnowledge.updatePerformance();
		
		// For some reason, this instruction writes KFs.dat for step=1. Step=2 
		// is not written. Step>2 is written from performAction.
		updateKFdata();
		
		// A matrix is created, where the actions undertaken by the firms are 
		// recorded. This information is needed by the graph of historic actions.
		firms_actions=new ArrayList <FirmAction>();
		firms_historic_actions=new ArrayList <FirmHistoricAction>();
		firms_historic=new HashMap <Integer, HistoricFirm>();
	
	} // end of buildObjects

	
	
	
	
	/**
	 * Several actions are carried out:
	 * - Dead agents are removed from their lists;
	 * - The measure of performance is updated;
	 * - The histograms of the business elements are updated.
	 */
	public void performAction() {
		// At simulation end, writes networks. Once networks are written, erases 
		// past actions.
		if ((step) == stopsAt) {
			SaveAllActionsGraphEnd();
			SaveActionGraphEnd();
			SaveKnowledgeGraphEnd();
			SaveSimilarityGraphEnd();	
			firms_historic_actions.clear();
			firms_historic.clear();
		}
		// During the simulation, it writes networks at specified intervals. 
		// Once networks are written, erases past actions.
		else if(((step)%saveNetworkStep)==0) {
			SaveAllActionsGraph();
			SaveActionGraph();
			SaveKnowledgeGraph();
			SaveSimilarityGraph();
			firms_historic_actions.clear();
			firms_historic.clear();
		}
		
		// Dead firms are removed from their lists and substituted with new ones.
		// The initial size of these new firms is a random draw from a uniform 
		// distribution between the smallest and the largest amon existing firms.
		// Finds out the largest firm
		double maxS = 0.0;
		for (int i = 0; i < firmList.size(); i++) {
			Firm aFirm = (Firm) firmList.get(i);
			if (aFirm.isDead() == false && aFirm.getSize() > maxS)
				maxS = aFirm.getSize();
		}
		// Finds out the smallest firm.
		double minS = maxS;
		for (int i = 0; i < firmList.size(); i++) {
			Firm aFirm = (Firm) firmList.get(i);
			if (aFirm.isDead() == false && aFirm.getSize() < minS)
				minS = aFirm.getSize();
		}
		// Finds dead firms and substitutes them.
		for (int i = 0; i < firmList.size(); i++) {
			Firm aFirm = (Firm) firmList.get(i);
			if (aFirm.isDead()) {
				int ide = aFirm.getIdentity();
				double siz = powerLaw(minS, maxS, sizeDistrSlope);
				int xPa = aFirm.getX();
				int yPa = aFirm.getY();
				int clu = aFirm.getYourCluster();
				// The random numbers generator that must be passed on to firms.
				if (changeSeed == true) {
					Random rsg = new Random();  // the random seed generator
					seed = rsg.nextInt();
				}
				Random rng = new Random(seed);  // the random numbers generator
				// Either creates an Imitator or an Innovator to substitute the 
				// dead firm.
				if (aFirm instanceof Imitator) {
					Imitator aNewImitator = 
							new Imitator(this, ide,	siz, xPa, yPa, clu);
					aNewImitator.initializeKnowledge(rng);
					firmList.add(aNewImitator);
					if (aFirm.getYourCluster() > 0)
						clusteredImitatorDeaths++;
					else
						isolatedImitatorDeaths++;
				} else if (aFirm instanceof Innovator) {
					Innovator aNewInnovator = 
							new Innovator(this, ide, siz, xPa, yPa, clu);
					aNewInnovator.initializeKnowledge(rng);
					firmList.add(aNewInnovator);
					if (aFirm.getYourCluster() > 0)
						clusteredInnovatorDeaths++;
					else
						isolatedInnovatorDeaths++;
				}
				firmList.remove(aFirm);
			}
		}


		// Updates the measure of performance.
		commonKnowledge.updatePerformance();
		
		// Updates the output: time series.
		updateStats();
		
		// Updates the output: knowledge fields and betweeness centrality.
		// For some reason, this instruction is only effective from step=3.
		// Step=1 is written from buildObjects (updateKFdata() at the end of 
		// buildObjects).
		// Step=2 is not written.
		// Beware: This instruction causes nasty problems if the model is run 
		// for 1 step only.
		updateKFdata();

		// Updates the output: business element instances.
		updateBEIdistribution();

	}  // end of performAction

	
	
	
	
	
	/**
	 * The output file for all time series produced by this model 
	 * at any point in time (66 variables):
	 * - Minimum number of rivals for isolated and clustered innovators and 
	 * imitators (4);
	 * - Maximum number of rivals for isolated and clustered innovators and 
	 * imitators (4);
	 * - Average of number of rivals for isolated and clustered innovators and 
	 * imitators + inside cluster rivals (6);
	 * - Sum of vicarious exploitation for isolated and clustered innovators and 
	 * imitators (4);
	 * - Sum of vicarious exploration for isolated and clustered innovators and 
	 * imitators (4);
	 * - Sum of experiential exploitation for isolated and clustered innovators 
	 * (2);
	 * - Sum of experiential exploration for isolated and clustered innovators 
	 * (2);
	 * - Sum of exit for isolated and clustered innovators and imitators (4);
	 * - Average cognitive distance for isolated and clustered innovators and 
	 * imitators (4);
	 * - Average performance for isolated and clustered innovators and imitators 
	 * (4);
	 * - Minimum number of knowledge fields for isolated and clustered 
	 * innovators and imitators (4);
	 * - Maximum number of knowledge fields for isolated and clustered 
	 * innovators and imitators (4);
	 * - Average number of knowledge fields for isolated and clustered 
	 * innovators and imitators (4);
	 * - Minimum expertise for isolated and clustered innovators and imitators 
	 * (4);
	 * - Maximum expertise for isolated and clustered innovators and imitators 
	 * (4);
	 * - Average expertise of isolated and clustered innovators and imitators 
	 * (4);
	 * - Sum of deaths of isolated and clustered innovators and imitators (4).
	 */
	public void updateStats() {
		
		// Variables that enter stats.dat. They are not declared globally. 
		
		// Rivals.
		int minRivalsIsImi = firmList.size() - 1; // but
		if (numIsolatedImitators == 0)
			minRivalsIsImi = 0;
		int minRivalsClImi = firmList.size() - 1; // but
		if (numClusteredImitators == 0)
			minRivalsClImi = 0;
		int minRivalsIsInn = firmList.size() - 1; // but
		if (numIsolatedInnovators == 0)
			minRivalsIsInn = 0;
		int minRivalsClInn = firmList.size() - 1; // but
		if (numClusteredInnovators == 0)
			minRivalsClInn = 0;
		int maxRivalsIsImi = 0;
		int maxRivalsClImi = 0;
		int maxRivalsIsInn = 0;
		int maxRivalsClInn = 0;
		int sumRivalsIsImi = 0;
		double aveRivalsIsImi = 0.0;
		int sumRivalsClImi = 0;
		double aveRivalsClImi = 0.0;
		int sumRivalsIsInn = 0;
		double aveRivalsIsInn = 0.0;
		int sumRivalsClInn = 0;
		double aveRivalsClInn = 0.0;
		int sumRivalsClImiInside = 0;
		double aveRivalsClImiInside = 0.0;
		int sumRivalsClInnInside = 0;
		double aveRivalsClInnInside = 0.0;
		// Actions.
		int sumVicPloiIsImi = 0;
		int sumVicPloiClImi = 0;
		int sumVicPloiIsInn = 0;
		int sumVicPloiClInn = 0;
		int sumVicPlorIsImi = 0;
		int sumVicPlorClImi = 0;
		int sumVicPlorIsInn = 0;
		int sumVicPlorClInn = 0;
		int sumExpPloiIsInn = 0;
		int sumExpPloiClInn = 0;
		int sumExpPlorIsInn = 0;
		int sumExpPlorClInn = 0;
		int sumExitIsImi = 0;
		int sumExitClImi = 0;
		int sumExitIsInn = 0;
		int sumExitClInn = 0;
		// Cognitive Distance.
		double sumCD_IsImi = 0.0;
		double aveCD_IsImi = 0.0;
		double sumCD_ClImi = 0.0;
		double aveCD_ClImi = 0.0;
		double sumCD_IsInn = 0.0;
		double aveCD_IsInn = 0.0;
		double sumCD_ClInn = 0.0;
		double aveCD_ClInn = 0.0;
		// Performance.
		double sumPerformanceIsImi = 0.0;
		double avePerformanceIsImi = 0.0;
		double sumPerformanceClImi = 0.0;
		double avePerformanceClImi = 0.0;
		double sumPerformanceIsInn = 0.0;
		double avePerformanceIsInn = 0.0;
		double sumPerformanceClInn = 0.0;
		double avePerformanceClInn = 0.0;
		// Number of knowledge fields.
		int minNKF_IsImi = maxFieldsPerFirm; // but
		if (numIsolatedImitators == 0)
			minNKF_IsImi = 0;
		int minNKF_ClImi = maxFieldsPerFirm; // but
		if (numClusteredImitators == 0)
			minNKF_ClImi = 0;
		int minNKF_IsInn = maxFieldsPerFirm; // but
		if (numIsolatedInnovators == 0)
			minNKF_IsInn = 0;
		int minNKF_ClInn = maxFieldsPerFirm; // but
		if (numClusteredInnovators == 0)
			minNKF_ClInn = 0;
		int maxNKF_IsImi = 0;
		int maxNKF_ClImi = 0;
		int maxNKF_IsInn = 0;
		int maxNKF_ClInn = 0;
		int sumNKF_IsImi = 0;
		double aveNKF_IsImi = 0.0;
		int sumNKF_ClImi = 0;
		double aveNKF_ClImi = 0.0;
		int sumNKF_IsInn = 0;
		double aveNKF_IsInn = 0.0;
		int sumNKF_ClInn = 0;
		double aveNKF_ClInn = 0.0;
		// Expertise of knowledge fields.
		double minExpertiseIsImi = maxExpertise; // but
		if (numIsolatedImitators == 0)
			minExpertiseIsImi = 0;
		double minExpertiseClImi = maxExpertise; // but
		if (numClusteredImitators == 0)
			minExpertiseClImi = 0;
		double minExpertiseIsInn = maxExpertise; // but
		if (numIsolatedInnovators == 0)
			minExpertiseIsInn = 0;
		double minExpertiseClInn = maxExpertise; // but
		if (numClusteredInnovators == 0)
			minExpertiseClInn = 0;
		double maxExpertiseIsImi = 0.0;
		double maxExpertiseClImi = 0.0;
		double maxExpertiseIsInn = 0.0;
		double maxExpertiseClInn = 0.0;
		double sumExpertiseIsImi = 0.0;
		double aveExpertiseIsImi = 0.0;
		double sumExpertiseClImi = 0.0;
		double aveExpertiseClImi = 0.0;
		double sumExpertiseIsInn = 0.0;
		double aveExpertiseIsInn = 0.0;
		double sumExpertiseClInn = 0.0;
		double aveExpertiseClInn = 0.0;
		
		// Scans the list of firms collecting information.
		for (int i=0; i<firmList.size(); i++) {
			Firm aFirm = firmList.get(i);
			// If it's an isolated firm.
			if (aFirm.getYourCluster() == 0) {
				// If it's an isolated Imitator.
				if (aFirm instanceof Imitator) {
					if (aFirm.getNumRivals() < minRivalsIsImi)
						minRivalsIsImi = aFirm.getNumRivals();
					if (aFirm.getNumRivals() > maxRivalsIsImi)
						maxRivalsIsImi = aFirm.getNumRivals();
					sumRivalsIsImi += aFirm.getNumRivals();
					sumVicPloiIsImi += aFirm.getVicariousExploitation();
					sumVicPlorIsImi += aFirm.getVicariousExploration();
					sumExitIsImi += aFirm.getExitField();
					sumCD_IsImi += aFirm.getCognitiveDistance();
					sumPerformanceIsImi += aFirm.getPerformance();
					if (aFirm.getNumFields() < minNKF_IsImi)
						minNKF_IsImi = aFirm.getNumFields();
					if (aFirm.getNumFields() > maxNKF_IsImi)
						maxNKF_IsImi = aFirm.getNumFields();
					sumNKF_IsImi += aFirm.getNumFields();
					if (aFirm.getLowestExpertise() < minExpertiseIsImi)
						minExpertiseIsImi = aFirm.getLowestExpertise();
					if (aFirm.getHighestExpertise() > maxExpertiseIsImi)
						maxExpertiseIsImi = aFirm.getHighestExpertise();
					sumExpertiseIsImi += aFirm.getAverageExpertise();
				}
				// If it's an isolated Innovator.
				else {
					if (aFirm.getNumRivals() < minRivalsIsInn)
						minRivalsIsInn = aFirm.getNumRivals();
					if (aFirm.getNumRivals() > maxRivalsIsInn)
						maxRivalsIsInn = aFirm.getNumRivals();
					sumRivalsIsInn += aFirm.getNumRivals();
					sumVicPloiIsInn += aFirm.getVicariousExploitation();
					sumVicPlorIsInn += aFirm.getVicariousExploration();
					sumExpPloiIsInn += aFirm.getExperientialExploitation();
					sumExpPlorIsInn += aFirm.getExperientialExploration();
					sumExitIsInn += aFirm.getExitField();
					sumCD_IsInn += aFirm.getCognitiveDistance();
					sumPerformanceIsInn += aFirm.getPerformance();
					if (aFirm.getNumFields() < minNKF_IsInn)
						minNKF_IsInn = aFirm.getNumFields();
					if (aFirm.getNumFields() > maxNKF_IsInn)
						maxNKF_IsInn = aFirm.getNumFields();
					sumNKF_IsInn += aFirm.getNumFields();
					if (aFirm.getLowestExpertise() < minExpertiseIsInn)
						minExpertiseIsInn = aFirm.getLowestExpertise();
					if (aFirm.getHighestExpertise() > maxExpertiseIsInn)
						maxExpertiseIsInn = aFirm.getHighestExpertise();
					sumExpertiseIsInn += aFirm.getAverageExpertise();
				}
			}
			// If it's a clustered firm.
			else {
				// If it's a clustered Imitator.
				if (aFirm instanceof Imitator) {
					if (aFirm.getNumRivals() < minRivalsClImi)
						minRivalsClImi = aFirm.getNumRivals();
					if (aFirm.getNumRivals() > maxRivalsClImi)
						maxRivalsClImi = aFirm.getNumRivals();
					sumRivalsClImi += aFirm.getNumRivals();
					sumRivalsClImiInside += aFirm.getNumRivalsInsideCluster();
					sumVicPloiClImi += aFirm.getVicariousExploitation();
					sumVicPlorClImi += aFirm.getVicariousExploration();
					sumExitClImi += aFirm.getExitField();
					sumCD_ClImi += aFirm.getCognitiveDistance();
					sumPerformanceClImi += aFirm.getPerformance();
					if (aFirm.getNumFields() < minNKF_ClImi)
						minNKF_ClImi = aFirm.getNumFields();
					if (aFirm.getNumFields() > maxNKF_ClImi)
						maxNKF_ClImi = aFirm.getNumFields();
					sumNKF_ClImi += aFirm.getNumFields();
					if (aFirm.getLowestExpertise() < minExpertiseClImi)
						minExpertiseClImi = aFirm.getLowestExpertise();
					if (aFirm.getHighestExpertise() > maxExpertiseClImi)
						maxExpertiseClImi = aFirm.getHighestExpertise();
					sumExpertiseClImi += aFirm.getAverageExpertise();
				}
				// If it's a clustered Innovator.
				else {
					if (aFirm.getNumRivals() < minRivalsClInn)
						minRivalsClInn = aFirm.getNumRivals();
					if (aFirm.getNumRivals() > maxRivalsClInn)
						maxRivalsClInn = aFirm.getNumRivals();
					sumRivalsClInn += aFirm.getNumRivals();
					sumRivalsClInnInside += aFirm.getNumRivalsInsideCluster();
					sumVicPloiClInn += aFirm.getVicariousExploitation();
					sumVicPlorClInn += aFirm.getVicariousExploration();
					sumExpPloiClInn += aFirm.getExperientialExploitation();
					sumExpPlorClInn += aFirm.getExperientialExploration();
					sumExitClInn += aFirm.getExitField();
					sumCD_ClInn += aFirm.getCognitiveDistance();
					sumPerformanceClInn += aFirm.getPerformance();
					if (aFirm.getNumFields() < minNKF_ClInn)
						minNKF_ClInn = aFirm.getNumFields();
					if (aFirm.getNumFields() > maxNKF_ClInn)
						maxNKF_ClInn = aFirm.getNumFields();
					sumNKF_ClInn += aFirm.getNumFields();
					if (aFirm.getLowestExpertise() < minExpertiseClInn)
						minExpertiseClInn = aFirm.getLowestExpertise();
					if (aFirm.getHighestExpertise() > maxExpertiseClInn)
						maxExpertiseClInn = aFirm.getHighestExpertise();
					sumExpertiseClInn += aFirm.getAverageExpertise();
				}
			}
		}
		
		// Computes averages over the number of Innovators and Imitators, isolated and clustered.
		if (numIsolatedImitators > 0) {
			aveRivalsIsImi = sumRivalsIsImi / numIsolatedImitators;
			aveCD_IsImi = sumCD_IsImi / numIsolatedImitators;
			avePerformanceIsImi = sumPerformanceIsImi / numIsolatedImitators;
			aveNKF_IsImi = sumNKF_IsImi / numIsolatedImitators;
			aveExpertiseIsImi = sumExpertiseIsImi / numIsolatedImitators;
		}
		if (numIsolatedInnovators > 0) {
			aveRivalsIsInn = sumRivalsIsInn / numIsolatedInnovators;
			aveCD_IsInn = sumCD_IsInn / numIsolatedInnovators;
			avePerformanceIsInn = sumPerformanceIsInn / numIsolatedInnovators;
			aveNKF_IsInn = sumNKF_IsInn / numIsolatedInnovators;
			aveExpertiseIsInn = sumExpertiseIsInn / numIsolatedInnovators;
		}
		if (numClusteredImitators > 0) {
			aveRivalsClImi = sumRivalsClImi / numClusteredImitators;
			aveRivalsClImiInside = sumRivalsClImiInside / numClusteredImitators;
			aveCD_ClImi = sumCD_ClImi / numClusteredImitators;
			avePerformanceClImi = sumPerformanceClImi / numClusteredImitators;
			aveNKF_ClImi = sumNKF_ClImi / numClusteredImitators;
			aveExpertiseClImi = sumExpertiseClImi / numClusteredImitators;
		}
		if (numClusteredInnovators > 0) {
			aveRivalsClInn = sumRivalsClInn / numClusteredInnovators;
			aveRivalsClInnInside = sumRivalsClInnInside / numClusteredInnovators;
			aveCD_ClInn = sumCD_ClInn / numClusteredInnovators;
			avePerformanceClInn = sumPerformanceClInn / numClusteredInnovators;
			aveNKF_ClInn = sumNKF_ClInn / numClusteredInnovators;
			aveExpertiseClInn = sumExpertiseClInn / numClusteredInnovators;
		}
		
		
		
		// Writes the variables on the file stats.
		try {
			if (stats != null)
				stats.write(
						(step) + " " + 
						minRivalsIsImi + " " + minRivalsIsInn + 
						" " + minRivalsClImi + " " + minRivalsClInn + " " +
						maxRivalsIsImi + " " + maxRivalsIsInn + 
						" " + maxRivalsClImi + " " + maxRivalsClInn + " " +
						aveRivalsIsImi + " " + aveRivalsIsInn + 
						" " + aveRivalsClImi + " " + aveRivalsClInn + 
						" " + aveRivalsClImiInside + 
						" " + aveRivalsClInnInside + " " +
						sumVicPloiIsImi + " " + sumVicPloiClImi + 
						" " + sumVicPloiIsInn + " " + sumVicPloiClInn + " " + 
						sumVicPlorIsImi + " " + sumVicPlorClImi + 
						" " + sumVicPlorIsInn + " " + sumVicPlorClInn + " " + 
						sumExpPloiIsInn + " " + sumExpPloiClInn + 
						" " + 
						sumExpPlorIsInn + " " + sumExpPlorClInn + 
						" " + 
						sumExitIsImi + " " + sumExitClImi + 
						" " + sumExitIsInn + " " + sumExitClInn + " " + 
						aveCD_IsImi + " " + aveCD_ClImi + 
						" " + aveCD_IsInn + " " + aveCD_ClInn + " " + 
						avePerformanceIsImi + " " + avePerformanceClImi + 
						" " + avePerformanceIsInn + " " + avePerformanceClInn + 
						" " + 
						minNKF_IsImi + " " + minNKF_IsInn + " " + minNKF_ClImi + 
						" " + minNKF_ClInn + " " +
						maxNKF_IsImi + " " + maxNKF_IsInn + " " + maxNKF_ClImi + 
						" " + maxNKF_ClInn + " " +
						aveNKF_IsImi + " " + aveNKF_IsInn + " " + aveNKF_ClImi + 
						" " + aveNKF_ClInn + " " +
						minExpertiseIsImi + " " + minExpertiseIsInn + 
						" " + minExpertiseClImi + " " + minExpertiseClInn + " " +
						maxExpertiseIsImi + " " + maxExpertiseIsInn + 
						" " + maxExpertiseClImi + " " + maxExpertiseClInn + " " +
						aveExpertiseIsImi + " " + aveExpertiseIsInn + 
						" " + aveExpertiseClImi + " " + aveExpertiseClInn + 
						" " + isolatedImitatorDeaths + 
						" " + clusteredImitatorDeaths + 
						" " + isolatedInnovatorDeaths + 
						" " + clusteredInnovatorDeaths
						+ "\r\n");
			stats.flush();
		} catch (Exception ee) {
			System.out.println("\n stats write print error: " + ee);
		}

		
		// Show line charts while the simulation is running.
				String stepNumber = ("step "+Integer.toString(step));
				
				datasetRivals.addValue( minRivalsIsImi , "minRivalsIsImi" , stepNumber );
				datasetRivals.addValue( minRivalsIsInn , "minRivalsIsInn" , stepNumber );
				datasetRivals.addValue( minRivalsClImi , "minRivalsClImi" , stepNumber );
				datasetRivals.addValue( minRivalsClInn , "minRivalsClInn" , stepNumber );
				datasetRivals.addValue( maxRivalsIsImi , "maxRivalsIsImi" , stepNumber );
				datasetRivals.addValue( maxRivalsIsInn , "maxRivalsIsInn" , stepNumber );
				datasetRivals.addValue( maxRivalsClImi , "maxRivalsClImi" , stepNumber );
				datasetRivals.addValue( maxRivalsClInn , "maxRivalsClInn" , stepNumber );
				datasetRivals.addValue( aveRivalsIsImi , "aveRivalsIsImi" , stepNumber );
				datasetRivals.addValue( aveRivalsIsInn , "aveRivalsIsInn" , stepNumber );
				datasetRivals.addValue( aveRivalsClImi , "aveRivalsClImi" , stepNumber );
				datasetRivals.addValue( aveRivalsClInn , "aveRivalsClInn" , stepNumber );
				datasetRivals.addValue( aveRivalsClImiInside , "aveRivalsClImiInside" , stepNumber );
				datasetRivals.addValue( aveRivalsClInnInside, "aveRivalsClInnInside" , stepNumber );
				
				datasetActions.addValue( sumVicPloiIsImi , "sumVicPloiIsImi" , stepNumber );
				datasetActions.addValue( sumVicPloiClImi , "sumVicPloiClImi" , stepNumber );
				datasetActions.addValue( sumVicPloiIsInn , "sumVicPloiIsInn" , stepNumber );
				datasetActions.addValue( sumVicPloiClInn , "sumVicPloiClInn" , stepNumber );
				datasetActions.addValue( sumVicPlorIsImi , "sumVicPlorIsImi" , stepNumber );
				datasetActions.addValue( sumVicPlorClImi , "sumVicPlorClImi" , stepNumber );
				datasetActions.addValue( sumVicPlorIsInn , "sumVicPlorIsInn" , stepNumber );
				datasetActions.addValue( sumVicPlorClInn , "sumVicPlorClInn" , stepNumber );
				datasetActions.addValue( sumExpPloiIsInn , "sumExpPloiIsInn" , stepNumber );
				datasetActions.addValue( sumExpPloiClInn , "sumExpPloiClInn" , stepNumber );
				datasetActions.addValue( sumExpPlorIsInn , "sumExpPlorIsInn" , stepNumber );
				datasetActions.addValue( sumExpPlorClInn , "sumExpPlorClInn" , stepNumber );
				datasetActions.addValue( sumExitIsImi , "sumExitIsImi" , stepNumber );
				datasetActions.addValue( sumExitClImi , "sumExitClImi" , stepNumber );
				datasetActions.addValue( sumExitIsInn , "sumExitIsInn" , stepNumber );
				datasetActions.addValue( sumExitClInn , "sumExitClInn" , stepNumber );
				
				datasetCognitiveDistance.addValue( aveCD_IsImi , "aveCD_IsImi" , stepNumber );
				datasetCognitiveDistance.addValue( aveCD_ClImi , "aveCD_ClImi" , stepNumber );
				datasetCognitiveDistance.addValue( aveCD_IsInn , "aveCD_IsInn" , stepNumber );
				datasetCognitiveDistance.addValue( aveCD_ClInn , "aveCD_ClInn" , stepNumber );
				
				datasetPerformance.addValue( avePerformanceIsImi , "avePerformanceIsImi" , stepNumber );
				datasetPerformance.addValue( avePerformanceClImi , "avePerformanceClImi" , stepNumber );
				datasetPerformance.addValue( avePerformanceIsInn , "avePerformanceIsInn" , stepNumber );
				datasetPerformance.addValue( avePerformanceClInn , "avePerformanceClInn" , stepNumber );
				
				datasetKnowledgeFields.addValue( minNKF_IsImi , "minNKF_IsImi" , stepNumber );
				datasetKnowledgeFields.addValue( minNKF_IsInn , "minNKF_IsInn" , stepNumber );
				datasetKnowledgeFields.addValue( minNKF_ClImi , "minNKF_ClImi" , stepNumber );
				datasetKnowledgeFields.addValue( minNKF_ClInn , "minNKF_ClInn" , stepNumber );
				datasetKnowledgeFields.addValue( maxNKF_IsImi , "maxNKF_IsImi" , stepNumber );
				datasetKnowledgeFields.addValue( maxNKF_IsInn , "maxNKF_IsInn" , stepNumber );
				datasetKnowledgeFields.addValue( maxNKF_ClImi , "maxNKF_ClImi" , stepNumber );
				datasetKnowledgeFields.addValue( maxNKF_ClInn , "maxNKF_ClInn" , stepNumber );
				datasetKnowledgeFields.addValue( aveNKF_IsImi , "aveNKF_IsImi" , stepNumber );
				datasetKnowledgeFields.addValue( aveNKF_IsInn , "aveNKF_IsInn" , stepNumber );
				datasetKnowledgeFields.addValue( aveNKF_ClImi , "aveNKF_ClImi" , stepNumber );
				datasetKnowledgeFields.addValue( aveNKF_ClInn , "aveNKF_ClInn" , stepNumber );
				
				datasetExpertise.addValue( minExpertiseIsImi , "minExpertiseIsImi" , stepNumber );
				datasetExpertise.addValue( minExpertiseIsInn , "minExpertiseIsInn" , stepNumber );
				datasetExpertise.addValue( minExpertiseClImi , "minExpertiseClImi" , stepNumber );
				datasetExpertise.addValue( minExpertiseClInn , "minExpertiseClInn" , stepNumber );
				datasetExpertise.addValue( maxExpertiseIsImi , "maxExpertiseIsImi" , stepNumber );
				datasetExpertise.addValue( maxExpertiseIsInn , "maxExpertiseIsInn" , stepNumber );
				datasetExpertise.addValue( maxExpertiseClImi , "maxExpertiseClImi" , stepNumber );
				datasetExpertise.addValue( maxExpertiseClInn , "maxExpertiseClInn" , stepNumber );
				datasetExpertise.addValue( aveExpertiseIsImi , "aveExpertiseIsImi" , stepNumber );
				datasetExpertise.addValue( aveExpertiseIsInn , "aveExpertiseIsInn" , stepNumber );
				datasetExpertise.addValue( aveExpertiseClImi , "aveExpertiseClImi" , stepNumber );
				datasetExpertise.addValue( aveExpertiseClInn , "aveExpertiseClInn" , stepNumber );

				datasetDeaths.addValue( isolatedImitatorDeaths , "isolatedImitatorDeaths" , stepNumber );
				datasetDeaths.addValue( clusteredImitatorDeaths , "clusteredImitatorDeaths" , stepNumber );
				datasetDeaths.addValue( isolatedInnovatorDeaths , "isolatedInnovatorDeaths" , stepNumber );
				datasetDeaths.addValue( clusteredInnovatorDeaths , "clusteredInnovatorDeaths" , stepNumber );
			    
				
				// display the line charts when achieving the last step
				if (step==stopsAt)
				{
			      LineChart chart = new LineChart("RIVALS" ,"RIVALS VALUES", datasetRivals);
			      chart.displayLineChart("RIVALS" ,"RIVALS VALUES", datasetRivals);
			      chart = new LineChart("ACTIONS" ,"ACTIONS VALUES", datasetActions);
			      chart.displayLineChart("ACTIONS" ,"ACTIONS VALUES", datasetActions);
			      chart = new LineChart("CognitiveDistance" ,"CognitiveDistance VALUES", datasetCognitiveDistance);
			      chart.displayLineChart("CognitiveDistance" ,"CognitiveDistance VALUES", datasetCognitiveDistance);
			      chart = new LineChart("PERFORMANCE" ,"PERFORMANCE VALUES", datasetPerformance);
			      chart.displayLineChart("PERFORMANCE" ,"PERFORMANCE VALUES", datasetPerformance);
			      chart = new LineChart("KnowledgeFields" ,"KnowledgeFields VALUES", datasetKnowledgeFields);
			      chart.displayLineChart("KnowledgeFields" ,"KnowledgeFields VALUES", datasetKnowledgeFields);
			      chart = new LineChart("datasetExpertise" ,"datasetExpertise VALUES", datasetExpertise);
			      chart.displayLineChart("datasetExpertise" ,"datasetExpertise VALUES", datasetExpertise);
			      chart = new LineChart("datasetDeaths" ,"datasetDeaths VALUES", datasetDeaths);
			      chart.displayLineChart("datasetDeaths" ,"datasetDeaths VALUES", datasetDeaths);
				}
		
		
		// Closes the output file at simulation end.
		try {
			if (step == stopsAt && run == simulation_number)
				stats.close();
		} catch (Exception eeee) {
			System.out.println("\n stats write print error: " + eeee);
		}

	}
	
	/**
	 * Output data on knowledge fields:
	 * - Simulation step;
	 * - What knowledge fields are there;
	 * - How many instances of each knowledge field exist;
	 * - A measure of performance (e.g., betweeness centrality) of each 
	 * knowledge field.
	 */
	private void updateKFdata() {
		for (KnowledgeField j : commonKnowledge.getKnowledgeFields().keySet()) {
			long identity = j.getID();
			int howMany = commonKnowledge.getNumInstancesOfField(j);
			double performance = commonKnowledge.getPerformance(j);
			// Writes a line of data.
			try {
				if (KFs != null)
					KFs.write(step + " " + identity + " " + howMany + 
							" " + performance + "\r\n");
			} catch (Exception ee) {
				System.out.println("\nGraphFile write print error: " + ee);
			}

		}
		
		// Closes the output file at simulation end.
		try {
			if (step == stopsAt)
				KFs.close();
		} catch (Exception eee) {
			System.out.println("\nGraphFile write print error: " + eee);
		}

	}
	
	/**
	 * The distribution of business element instances.
	 * This prints the following information on BEIs.dat:
	 * - Simulation step;
	 * - Business element starting from 1, with 0 representing undefined business elements;
	 * - The name of each business element instance;
	 * - The number of replicas;
	 * - The number of firms that have at least one replica.
	 */
	private void updateBEIdistribution() {
		
		for (int j=0; j<numBusinessElements; j++) {
			ArrayList<BusinessElementInstance> l = instances[j];
			for (BusinessElementInstance abei : l) {
				// The business element, where 0 represents undefined business 
				// elements.
				int be = 0;
				if (abei.isAspectUndefined() == false)
					be = j + 1;
				// The ID of this business element instance.
				long name = abei.getInstanceOfAspect();
				// Counts the replicas of this business element instance.
				int count= 0; // how many replicas of this BEI
				// Counts the number of firms that have this business element 
				// instance.
				int firmsWithThisBEI = 0;
				for (int i = 0; i < firmList.size(); i++) {
					Firm aFirm = (Firm) firmList.get(i);
					KnowledgeBase itsKnowledge = aFirm.getKnowledgeBase();
					count += itsKnowledge.howManyOfThisBEI(name);
					if (itsKnowledge.howManyOfThisBEI(name) > 0)
						firmsWithThisBEI++;
				}
				// If at least 1 firm has this bei, print information about it.
				if (firmsWithThisBEI > 0) {
					try {
						if (BEIs != null)
							BEIs.write(step + " " + be + " " + name + 
									" " + count + " " + firmsWithThisBEI + 
									"\r\n");
					} catch (Exception ee) {
						System.out.println("\nGraphFile write print error: " + ee);
					}	

				}
			}
		}
		
		// Closes the output file at simulation end.
		try {
			if (step == stopsAt)
				BEIs.close();
		} catch (Exception eee) {
			System.out.println("\nGraphFile write print error: " + eee);
		}
		
	}
	
	/**
	 * The main. Here is where everything starts.
	 *
	 * @param args
	 */
	public static void main(String[] args) {
		// Several simulations may be run. This makes sense if changeSeed==true.
		for (run=1; run<=simulation_number; run++) {
			//	Main builds Model. Model builds objects. Initialization.
			System.out.println("Run " + run);
			Model model = new Model();
			model.buildObjects();
			// The normal operation of the model, step-by-step.
			for (step=1; step<=stopsAt; step++) {
				System.out.println("Step " + step);
				for (int i = 0; i < firmList.size(); i++) {
					Firm aFirm = firmList.get(i);
					aFirm.performAction();
				}
				model.performAction();
			}
		}
	}
	
	/**
	 * MATHEMATICAL FUNCTIONS
	 */

	/**
	 *  This function is called by buildObjects in order to initialize the size 
	 *  of firms according to a power law.
	 *  It is also called in order to substitute dead firms.
	 *  
	 * @param min
	 * @param max
	 * @param exponent
	 * @return
	 */
	public double powerLaw(double min, double max, double exponent) {
		Random rng = new Random(seed);  // this is used to generate a uniform 
										// distribution, subsequently turned 
										// into a power law
		double uniform = min + (max - min)*rng.nextDouble();  // the uniform 
										// distribution in (0, 1) is translated 
										// into (min, max)
		return Math.pow(((Math.pow(max, (exponent+1)) - 
				Math.pow(min, (exponent+1)))*uniform + 
				Math.pow(min, (exponent+1))), 1/(exponent+1));
	}
	
	/**
	 *  This function returns a coordinate on a torus worldLength x worldLength.
	 *  
	 * @param c
	 * @return
	 */
	public int getTorus(int c) {
		return (c % worldLength + worldLength) % worldLength;
	}	
}  // the end
