
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Random;


/*!
 * The parent class of all firms.
 * - 1. Rivals are identified;
 * - 2. Knowledge is updated;
 * - 3. Conditions for bankruptcy are checked.
 * 
 *  \author Sirio Capizzi
 *  \author Guido Fioretti
 *  \author Ruggero Rossi
 *  \copyright GNU Public License.
 */
public abstract class Firm {
	
	// Objects in the environment.
	protected Model model;	///< The model that creates the agents.
	protected CommonKnowledge commonKnowledge;	///< Knowledge fields are common 
												///< knowledge.
	
	// Parameters regarding the identification of rivals.
	protected boolean rivalsChanged;	///< Whether the set of rivals changed.
	protected double similarityThreshold;	///< Other firms are considered rivals if 
											///< the similarity of their 
											///< knowledge is greater than this 
											///< threshold
	protected double maxCognitiveEffort;	///< The maximum cognitive effort to 
											///< be spent in order to take 
											///< account of rivals
	protected ArrayList<Firm> firmList;	///< The list of firms, from which the 
										///< rivals are drawn.
	protected double valuesSI[];	///< Firms' similarity index: a vector of 
									///<firmList.size()
	protected ArrayList<Firm> rivalsList = new ArrayList<Firm>();  // This firm + its rivals
	// (includes this firm because this is required in order to draw Action graphs)
	
	// Parameters regarding the firm's behavior.
	protected int identity;  ///< A firm's identity.
	protected double size;  ///< A firm's size, the integral of performance.
	protected double sizeAttained;  ///< The maximum size attained, needed in 
									///< order to implement linear decay.
	protected int xPos, yPos;  ///< A firm's position.
	protected int cluster;  ///< The firm is a member of this cluster.
	protected double avePerformance;	///< Performance averaged over all 
										///< knowledge fields.
	protected double lowPerformance;	///< The lower threshold of performance 
										///< as a percentage of the time average 
										///< of past performance.
	protected double minCriticalPerformance;	///< The lower threshold of 
												///< performance, adjusted to 
												///< past values.
	protected boolean dead = false;	///< Dead firms will be eliminated.
	
	// Objects and parameters regarding the agent's knowledge.
	protected KnowledgeBase myKnowledge;	///< A firm's knowledge base.
	protected KnowledgeBase rivalKnowledge;	///< A rivals' knowledge base, or 
											///< the compound knowledge of rivals.
	protected double aveCognitiveDistance;  ///< Cognitive distance averaged 
											///< over all knowledge fields.
	protected double minCriticalCognitiveDistance;	///< The lower threshold of 
													///< cognitive distance, 
													///< adjusted to past values.
	protected double maxCriticalCognitiveDistance;  ///< The upper threshold of 
													///< cognitive distance, 
													///< adjusted to past values.
	protected int maxNumFields;	///< The maximum number of fields.
	protected double maxExpertise;	///< The maximum expertise that can be 
									///< attained.
	protected double decayRate;	///< The rate of decay of capital and knowledge.
	
	// The indicators and their visibility.
	protected int expExploration; 	///< The number of times an original 
									///< Exploration of new knowledge.
	protected int vicExploration;	///< The number of times new knowledge is 
									///< explored by imitation.
	protected int expExploitation;  ///< The number of times existing knowledge 
									///< is exploited.
	protected int vicExploitation;  ///< The number of times existing knowledge 
									///< is exploited by imitation.
	protected int exit;	///< The number of times a firm abandons a knowledge 
						///< field.
	public double numRivalsInsideCluster;	///< The number of rivals in one's 
											///< own cluster (meaningful for 
											///< clustered firms only).
	protected double numRivals;	///< The total number of rivals.
	protected int numFields;	///< The number of knowledge fields of a firm.
	protected double lowestExpertise;	///< The lowest expertise among 
										///< knowledge fields.
	protected double averageExpertise;  ///< The average expertise of knowledge 
										///< fields.
	protected double highestExpertise;  ///< The highest expertise among 
										///< knowledge fields.
	
	int areMostRivalsFar = -1;
	
	/**
	 * 
	 * @param m
	 * @param id
	 * @param s
	 * @param x
	 * @param y
	 * @param c
	 */
	public Firm(Model m, int id, double s, int x, int y, int c) {
		model = m;
		identity = id;  // the identity of this firm
		size = s;  // the size of this firm
		sizeAttained = size;  // the maximum size it will attain is initialized
		xPos = x;  // its position on the x-axis
		yPos = y;  // its position on the y-axis
		cluster = c;  // it is a member of this cluster
		firmList = model.firmList;  // the list of all firms
		// A vector entailing values of the Similarity Index with all other firms, initialized to zero.
		valuesSI = new double[model.numClusteredInnovators + model.numClusteredImitators +
		                      model.numIsolatedInnovators + model.numIsolatedImitators];
		for (int i=0; i<(model.numClusteredImitators + model.numClusteredInnovators +
				model.numIsolatedImitators + model.numIsolatedInnovators); i++)
			valuesSI[i] = 0.0;
		maxCognitiveEffort = model.maxCognitiveEffort;  // the maximum cognitive 
											// effort to be spent in 
											//order to take account of rivals
		maxNumFields = model.maxFieldsPerFirm;  // the maximum number of 
											// knowledge fields for each firm
		maxExpertise = model.maxExpertise;  // the maximum depth that can be 
											// attained
		myKnowledge = new KnowledgeBase(maxExpertise);
		rivalKnowledge = new KnowledgeBase(maxExpertise);
		commonKnowledge = model.commonKnowledge;  	// the instances of all 
													// knowledge fields
		minCriticalPerformance = 0.0;  	// initialization of the lower threshold 
										// of performance
		minCriticalCognitiveDistance = 0.5 - model.cognitiveDistanceWindow / 2;  
			// the lower threshold of cognitive distance
		maxCriticalCognitiveDistance = 0.5 + model.cognitiveDistanceWindow / 2;  
			// the upper threshold of cognitive distance
		decayRate = model.decayRate;  
			// the rate of decay of capital and knowledge
	}
	
	/**
	 * Initialization of knowledge.
	 * A substantial part of initialization is carried out within buildObjects 
	 * of Model, where BEIs are created.
	 * Here, a subset of these BEIs is picked up.
	 *
	 * @param RandomNumbersGenerator
	 */
	public void initializeKnowledge(Random RandomNumbersGenerator) {
		// If the seed is fixed, it is divided by each firm's identity so each 
		// firm has its own, fixed seed.
		// If it's not fixed, each firm generates a different random seed.		
		long localSeed;
		if (model.changeSeed == true) {
			Random rsg = new Random();  // the random seed generator
			localSeed = rsg.nextInt();
		}
		else
			localSeed = model.seed/(identity + 1);  // each firm has a different seed
		Random rng = new Random(localSeed);  // the random numbers generator
	
		// It endows this firm with a number of knowledge fields which is either the maximum allowed,
		// or the maximum number of different knowledge fields that can be 
		// generated with the available BEIs.
		int availableBEIs = Model.varietyAtInitialization;
		int numBEs = Model.numBusinessElements;
		int possibleNumFields = (int)Math.pow(availableBEIs, numBEs); // the maximum number of kfs that can be generated
		// If the procedure picks up a knowledge field that is already there, it is not added to myKnowledge.
		// Thus, the number of knowledge fields can be smaller than Math.min(maxNumFields, possibleNumFields).
		for (int i=0; i < Math.min(maxNumFields, possibleNumFields); i++) {
			// Generates a new knowledge field.
			KnowledgeField kf=new KnowledgeField();	
			// Adds BEIs to it.
			for (int j=0; j<numBEs; j++) {
				ArrayList<BusinessElementInstance> l = Model.instances[j];
				int randomIndex = rng.nextInt(availableBEIs);
int aspect = l.get(randomIndex).getWhichAspect();
long bei = l.get(randomIndex).getInstanceOfAspect();
System.out.println("L'impresa " + identity + " riceve aspect " + aspect + " instance " + bei);
				kf.addBEI(l.get(randomIndex));				
			}
	
			// If the newly generated knowledge field is novel for this firm, 
			// it goes on.
			if (myKnowledge.isThereField(kf) == false) {  
				// Model creates BEIs with IDs that can be compared
				double initialExpertise = maxExpertise*rng.nextFloat();  
					// the initial expertise is obtained randomly
				myKnowledge.setField(kf, initialExpertise);  
					// a new knowledge field is created and its expertise is set
				commonKnowledge.updateField(kf, 1);	 
					// This changes a kf's ID if it is equivalent to an already 
					// existing one
			}
		}
System.out.println("Fine dell'inizializzazione della conoscenza da parte dell'impresa");
	}

	/**
	 * Updates the threshold of performance.
	 *
	 * @param avep
	 */
	public void updatePerformanceThreshold(double avep) {
		ArrayList<Double> pastAvePerformance = new ArrayList<Double>();  
			// past values of performance
		if (avep > 0)
			pastAvePerformance.add(avep);  	// stores in memory the most recent 
											// performance
		if (pastAvePerformance.size() > 10) do {
			pastAvePerformance.remove(0);  // forgets the oldest performance
		}
		while (pastAvePerformance.size() > 10);
		// Computes minCriticalPeformance
		minCriticalPerformance = 10000000000000.0;
		for (int i=0; i<pastAvePerformance.size(); i++) {
			if (pastAvePerformance.get(i) < minCriticalPerformance)
				minCriticalPerformance = pastAvePerformance.get(i);
		}
	}
	
	/**
	 * This method is called by the 'main', which is inside Model.
	 * In this case, performAction has been subdivided into:
	 * - checkAlive;
	 * - identifyRivals;
	 * - updateKnowledge.
	 */
	public void performAction()	{
		checkAlive();
		if (!isDead()) {  
			// only if it's alive the other operations are carried out
			identifyRivals();
			updateKnowledge();
		}
	}
	
	/**
	 * The firm checks that (1) it has sufficient financial resources to do 
	 * anything and that (2) it has at least one knowledge field left.
	 * If it has not, the boolean variable "dead" is set true.
	 * If it still has knowledge fields, they are deleted from CommonKnowledge. 
	 * The counter of exits is increased.
	 * Later on, Model kills the agents whose boolean variable "dead" is true.
	 */
	public void checkAlive()
	{	
		if ((size <= 0 || myKnowledge.getNumFields() == 0) && !dead) {
			dead = true;
			if (myKnowledge.getNumFields() > 0) {
			    for (KnowledgeField kf: myKnowledge.fields()) {
			    	commonKnowledge.updateField(kf, -1);
			    	exit++;
			    }
			}
		}	
	}
	
	/**
	 * 
	 * @return
	 */
	public boolean isDead() {
		return dead;
	}

	/**
	 * Rivals are selected among the most similar competitors, independently of 
	 * distance.
	 * However, the more distant the rivals, the more cognitive effort is 
	 * necessary to spend.
	 * The search for rivals stops as soon as the maximum cognitive effort 
	 * allowed is reached.
	 * This search is only carried out if similarity indexes changed.
	 */
	public void identifyRivals()
	{
		// This boolean variable is used to decide whether compoundKnowledge 
		// must be computed again.
		// More exactly, it only means that the values of the similarity index 
		// and, therefore, rivals are identified again.
		// Strictly speaking, it may happen that certain values of the 
		// similarity index change, rivals are identified again,
		// and yet they are the same as in the previous step.
		rivalsChanged = false;

		// Vector valuesSI entails the values of the similarity index for all 
		// firms in firmList.
		// It spares on the computation of the similarity index (SI) later on.
		// High values of SI mean that two firms have similar knowledge.
		// If SI is updated, rivalsChanged is set to true.
		for (int i=0; i<firmList.size(); i++) {
			KnowledgeBase otherKnowledge = firmList.get(i).myKnowledge;
			double SI = KnowledgeOperators.getSimilarityIndex(myKnowledge, 
					otherKnowledge);
			if (SI < (valuesSI[i] - 0.01) || SI > (valuesSI[i] + 0.01)) {
				valuesSI[i] = SI;
				if (rivalsChanged == false)
					rivalsChanged = true;
			}
		}

		// If the values of the similarity index changed, rivalsList is 
		// eventually updated.
		// Rivals are selected for high similarity index (SI).
		// However, the cognitive effort that is actually spent depends on 
		// physical distance.
		// Search for rivals stops as soon as cognitive effort reaches its cap,
		// or the list of firms has been searched as many times as its length.
		if (rivalsChanged == true) {
			rivalsList.clear();
			rivalsList.add(this); // aggiunta per evitare un ulteriore lista per scrivere i grafici
			double cognitiveEffort = 0.0;
			int counter = 0;
			do {
				Firm potentialRival = null;
				Firm newRival = null;
				double highestSI = 0.0;
				for (int i=0; i<firmList.size(); i++) {
					// Splits checking into 2 'if's in order not to check all 
					// conditions if SI is not at its peak.
					if ((valuesSI[i]) > (highestSI)) {
						potentialRival = firmList.get(i);
						if (potentialRival != null && 
								potentialRival != this && 
								!rivalsList.contains(potentialRival)) {
							newRival = potentialRival;
							highestSI = valuesSI[i];
						}
					}
				}
				counter++;
				if (newRival != null) {
					rivalsList.add(newRival);
					cognitiveEffort += 
							(1 - newRival.getDistanceCoefficient(xPos, yPos));
				}
			}
			while (cognitiveEffort < maxCognitiveEffort && 
					counter < firmList.size());
		}

		// Computes the indicators.
		numRivals = (double)rivalsList.size() - 1;  // because rivalsList entails this firm
		numRivalsInsideCluster = 0;
		for (int i=0; i<rivalsList.size(); i++) {
			if (cluster != 0) { // if this firm is in a cluster
				if (rivalsList.get(i) != this && (rivalsList.get(i)).getYourCluster() == cluster)
					numRivalsInsideCluster++;
			}
		}
		
	}
	
	/**
	 * Decisions that affect knowledge are carried out.
	 * This methods computes the data upon which experimental exploration, 
	 * experimental exploitation, vicarious exploration, vicarious exploitation, 
	 * and exit are carried out by the children.
	 * Exit is carried out here, after all other actions.
	 * Furthermore, this method measures:
	 * - The number of knowledge fields;
	 * - Minimum, maximum and average expertise;
	 * - Size and sizeAttained;
	 * - Average performance of this firm's knowledge fields;
	 * - Average cognitive distance of this firm's knowledge base from its rivals.
	 */
	public void updateKnowledge()
	{
		// Needed for graph initialization.
		if(!model.firms_actions.isEmpty())
		{
			FirmAction act=model.firms_actions.get(0);
			if(act.step!=model.step)
				model.firms_actions.clear();
		}
		
		// Initializes the counters.
		int pcounter = 0;  	// counts non-isolated knowledge fields
		int qcounter = 0;  	// counts rival's knowledge fields multiplied by own knowledge fields

		// Initializes the indicators.
		expExploration = 0;
		vicExploration = 0;
		expExploitation = 0;
		vicExploitation = 0;
		exit = 0;
		// Initializes the evaluators.
		double totalCognitiveDistance = 0.0f;	// the sum of cognitive distance 
											// over all pairs of knowledge fields
		double totalPerformance = 0.0;	// the sum of performance over all 
										// knowledge fields

		// Prepares knowledge maps.
		// Cognitive distance is computed with respect to rivalKnowledge.
		// If there is at least one rival, and if the list of rivals changed.
		if (rivalsList.size() > 1 && rivalsChanged) {
			// Initialization of rivalKnowledge to the first rival's knowledge.
			if (rivalKnowledge == null)
				System.out.println("rK e' null");
			/*try {
				if (rivalKnowledge != null)
					rivalKnowledge.removeAllFields();
			} catch (Exception e) {
				System.out.println("NULL rivalKnowledge: " + e);
			}*/
			//else
			rivalKnowledge.removeAllFields();
			int j = 0;
			do  {
				Firm rival = (Firm)rivalsList.get(j);
				if (rival != this) {
					rivalKnowledge = (KnowledgeBase)rival.getKnowledgeBase();
					double DC = rival.getDistanceCoefficient(xPos, yPos);  
					rivalKnowledge.accountForDistance(DC);
				}
				j++;
			} while (rivalKnowledge.isEmpty() && j < rivalsList.size());
				
			KnowledgeBase yourKnowledge;
			for (int i=0; i<rivalsList.size(); i++) {
				if (rivalsList.get(i) != this) { // rivalsList entails the calling agent, too
					Firm rival = (Firm)rivalsList.get(i);  // a rival is considered
					yourKnowledge = (KnowledgeBase)rival.getKnowledgeBase();
					// if it's neither null nor empty, yourKnowledge enters the compound rival's knowledge
					if (yourKnowledge != null && !yourKnowledge.isEmpty()) {
						// its distance coefficient from the calling firm
						double dc = rival.getDistanceCoefficient(xPos, yPos);  
						// the current rival's knowledge, weighted by distance
						yourKnowledge.accountForDistance(dc);  
						// If only one rival is in the list there is no need to call Dempster-Shafer's rule.
						// rivalKnowledge coincides with yourKnowledge
						if (rivalsList.size() == 2) // the one rival + this firm
							rivalKnowledge = yourKnowledge;
						// If at least two rivals are in the list, Dempster-Shafer's rule is eventually called
						// in order to combine yourKnowledge-s into a compound rivalKnowledge
						if (rivalsList.size() > 2) {
								rivalKnowledge = KnowledgeOperators.dempsterShafer(
										rivalKnowledge, yourKnowledge, commonKnowledge);
						}
					} 
				}  // ends if yourKnowledge is not null and not empty
			}  // ends scanning rivalsList
		}  // ends if at least one rival exist, and rivals changed

		// No action is undertaken if either my knowledge is null, or the 
		// rivals' compound knowledge is null.
		if (myKnowledge != null && rivalKnowledge != null) {
			// Prepares the map for scanning rivals' knowledge.
			ArrayList<Map.Entry<KnowledgeField,ExpertiseData>> rES = 
					new ArrayList<Map.Entry<KnowledgeField,ExpertiseData>>(
							rivalKnowledge.getKnowledge());
			// Prepares the map for scanning its own knowledge.
			ArrayList<Map.Entry<KnowledgeField,ExpertiseData>> myES = 
					new ArrayList<Map.Entry<KnowledgeField,ExpertiseData>>(
							myKnowledge.getKnowledge());
		
			// Scans own knowledge.
			for(Map.Entry<KnowledgeField,ExpertiseData> p: myES) {
				double performance;	// the performance of a single knowledge 
									// field
				KnowledgeField pk = p.getKey();
System.out.println("dentro updateknowledge, mio kf: " + pk + " " + pk.getInstance(0).getInstanceOfAspect() + " " + pk.getInstance(1).getInstanceOfAspect());
System.out.println("Questo kf contiene undefined BEIs: " + pk.entailsUndefinedBEI());
System.out.println("CommonKnowledge contiene questo kf: " + commonKnowledge.containsKF(pk));
System.out.println("La mia knowledge base contiene questo kf: " + myKnowledge.isThereField(pk));

double pv = p.getValue().expertise;
				
				// Some forgetting takes place. QUESTA E' L'ISTRUZIONE CHE GENERA ERRORE
				myKnowledge.tuneFieldExpertise(pk, -decayRate);  
				// Exit a knowledge field if its expertise is equal or less than zero.
				exitField(pk);
				
				// If there are rivals, cognitive distance is computed and 
				// vicarious exploitation is eventually carried out.
				// If there are no rivals, or if rivals have no field at low or 
				// intermediate cognitive distance.
				boolean anypAtLowIntermediateCognitiveDistanceWithSufficientExpertise = false;
				if (rivalsList.size() > 1) {
					// Scans rival's knowledge for every knowledge field of 
					// one's knowledge.
					for(Map.Entry<KnowledgeField,ExpertiseData> q: rES) {
						double cognitiveDistance;	// the cognitive distance for 
												// each pair of knowledge fields
						KnowledgeField qk = q.getKey();
						double qv = q.getValue().expertise;
						// Cognitive Distance is computed.
						double intersection = qk.intersectionValue(pk); 
							// the qualitative difference between the two 
							// knowledge fields
						cognitiveDistance = 1 - intersection;
						totalCognitiveDistance += cognitiveDistance;
						// this will be used in experiential learning.
						if (cognitiveDistance <= 
								maxCriticalCognitiveDistance && qv > pv)
							anypAtLowIntermediateCognitiveDistanceWithSufficientExpertise = true;
						qcounter++;
						// Eventually, vicarious exploitation is carried out.
						vicariousExploitation(pk, pv, qv, cognitiveDistance);
					}
				}
				
				// Performance is computed.
				performance = commonKnowledge.getPerformance(pk);
				if (performance > 0) {
					totalPerformance += performance;
					pcounter++;
				}
				// Eventually, experiential learning is carried out.
				experientialExploration(pk, performance, anypAtLowIntermediateCognitiveDistanceWithSufficientExpertise);
				experientialExploitation(pk, pv, performance, anypAtLowIntermediateCognitiveDistanceWithSufficientExpertise);
				
				// Some forgetting takes place. QUESTA E' L'ISTRUZIONE CHE GENERA ERRORE
				/*myKnowledge.tuneFieldExpertise(pk, -decayRate);  
				// Exit a knowledge field if its expertise is equal or less than zero.
				exitField(pk);*/
			}	 
	
			// Scans rival's knowledge.
			if (rivalsList.size() > 1) {
				for(Map.Entry<KnowledgeField,ExpertiseData> q: rES) {
					KnowledgeField qk = q.getKey();
					double qv = q.getValue().expertise;
					// The average cognitive distance for each pk.
					boolean anyqAtIntermediateCognitiveDistance = false;
					// Scans own knowledge for every knowledge field of rival's 
					// knowledge.
					for(Map.Entry<KnowledgeField,ExpertiseData> p: myES) {
						KnowledgeField pk = p.getKey();
						double intersection1 = qk.intersectionValue(pk); 
							// the qualitative difference between the two 
							// knowledge fields
						double cognitiveDistance1 = 1 - intersection1;
						if (cognitiveDistance1 >= minCriticalCognitiveDistance && 
								cognitiveDistance1 <= maxCriticalCognitiveDistance)
							anyqAtIntermediateCognitiveDistance = true;					
					}
					// Eventually, vicarious exploration is carried out.
					vicariousExploration(qk, qv, anyqAtIntermediateCognitiveDistance);
				}
			}
		} // end of 'if (myKnowledge != null && rivalKnowledge != null)'
		

		
		// Two indicators of absorptive capacity: The number of knowledge fields 
		// and the expertise of knowledge fields.
		numFields = myKnowledge.getNumFields();
		lowestExpertise = myKnowledge.getLowestExpertise();
		averageExpertise = myKnowledge.getAverageExpertise();
		highestExpertise = myKnowledge.getHighestExpertise();			

		// Size and sizeAttained are updated. Negative size is set to zero in 
		// order not to show it in the one step before death.
		size += (totalPerformance - decayRate*sizeAttained);	// size is the 
												// integral of performance, 
												// minus capital depreciation
		if (size > sizeAttained)
			sizeAttained = size;
		if (size < 0)
			size = 0;

		// Average performance is computed and communicated to the evaluator of 
		// performance threshold.
		if (totalPerformance > 0 && pcounter > 0)
			avePerformance = totalPerformance/pcounter;
		else
			avePerformance = 0.;
		updatePerformanceThreshold(avePerformance);
		
		// Average cognitive distance is computed.
		if (totalCognitiveDistance > 0 && qcounter > 0)
			aveCognitiveDistance = totalCognitiveDistance/qcounter;
		else
			aveCognitiveDistance = 0;
		

	}  // end of updateKnowledge
	
	/**
	 * These methods are called by "updateKnowledge".
	 * They are written in the children, because they specify the actions 
	 * undertaken by each kind of agent.
	 */
	
	/**
	 * 
	 * @param kf1
	 * @param p
	 * @param anyLowIntermediateCD
	 */
	public abstract void experientialExploration(KnowledgeField kf1, double p, 
			boolean anyLowIntermediateCD);
	
	/**
	 * 
	 * @param kf1
	 * @param currentExpertise
	 * @param p
	 * @param anyLowIntermediateCD
	 */
	public abstract void experientialExploitation(KnowledgeField kf1, 
			double currentExpertise, double p, boolean anyLowIntermediateCD);
	
	/**
	 * 
	 * @param kf2
	 * @param wd2
	 * @param anyIntermediateCD
	 */
	public abstract void vicariousExploration(KnowledgeField kf2, double wd2, 
			boolean anyIntermediateCD);
	
	/**
	 * 
	 * @param kf1
	 * @param kf2
	 * @param d1
	 * @param d2
	 * @param cd
	 */
	public abstract void vicariousExploitation(KnowledgeField kf1, double d1, double d2, double cd);
	
	/**
	 * 
	 * @param kf
	 */
	public abstract void exitField(KnowledgeField kf);
	
	/**
	 * These methods return a firm's actions. They are implemented in the 
	 * children.
	 */
	/**
	 * 
	 * @return
	 */
	public abstract int getExperientialExploration();
	
	/**
	 * 
	 * @return
	 */
	public abstract int getExperientialExploitation();
	
	/**
	 * 
	 * @return
	 */
	public abstract int getVicariousExploration();
	
	/**
	 * 
	 * @return
	 */
	public abstract int getVicariousExploitation();
	
	/**
	 * 
	 * @return
	 */
	public abstract int getExitField();
	
	/**
	 * This method is written in the children. It tells whether it's an 
	 * Innovator or an Imitator.
	 */
	public abstract java.awt.Color getColor();

	
	
	
	
	
	
	
	
	
	/**
	 * 
	 * @return Returns its identification number to other agents.
	 */
	public int getIdentity() {
		return identity;
	}
	
	/**
	 * 
	 * @return Returns the cluster it is part of.
	 */
	public int getYourCluster() {
		return cluster;
	}
	
	/**
	 * 
	 * @return These methods return the number of rivals.
	 */
	public int getNumRivals() {
		return (int)numRivals;
	}
	
	/**
	 * 
	 * @return These methods return the number of rivals.
	 */
	public int getNumRivalsInsideCluster() {
		return (int)numRivalsInsideCluster;
	}
	
	/**
	 * Returns the similarity index with another firm.
	 * 
	 * @return
	 */
	public double getValueSI(int n) {
		return valuesSI[n];
	}
	
	/**
	 * 
	 * @return Returns a firm's cognitive distance from its rivals, averaged 
	 * over all of its knowledge fields.
	 */
	public double getCognitiveDistance() {
		return aveCognitiveDistance;
	}
	
	/**
	 * Returns the number of knowledge fields.
	 * 
	 * @return
	 */
	public int getNumFields() {
		return numFields;
	}
	
	/** 
	 * Return lowest, highest and average depth of a firm's knowledge fields.
	 * 
	 * @return
	 */
	public double getLowestExpertise() {
		return lowestExpertise;
	}
	
	/**
	 * 
	 * @return
	 */
	public double getAverageExpertise() {
		return averageExpertise;
	}
	
	/**
	 * 
	 * @return
	 */
	public double getHighestExpertise() {
		return highestExpertise;
	}
	
	/**
	 * Returns its performance.
	 *
	 * @return Returns its performance.
	 */
	public double getPerformance() {
		return avePerformance;
	}
	
	/**
	 * Returns its size.
	 * 
	 * @return Returns its size.
	 */
	public double getSize() {
		return size;
	}
	
	/**
	 * Return its position.
	 * 
	 * @return Return its X position.
	 */
	public int getX() {
		return xPos;
	}
	
	/**
	 * Return its position.
	 * 
	 * @return Return its Y position.
	 */
	public int getY() {
		return yPos;
	}
	
	/**
	 * Returns a distance coefficient in the [0,1] interval: 0 denotes maximum 
	 * distance, 1 denotes zero distance. It assumes a torus.
	 * It refers to the distance between the quering firm and this firm.
	 * The distance coefficient is 1 when distance has no effect.
	 *
	 * @param x
	 * @param y
	 * @return
	 */
	public double getDistanceCoefficient(int x, int y) {
		double distance;
		double coefficient;
		double worldDiagonal = Math.sqrt(Math.pow(model.worldLength, 2) + 
				Math.pow(model.worldLength, 2));
		distance = Math.sqrt(Math.pow(Math.min(Math.abs(x - xPos), 
				model.worldLength - Math.abs(x - xPos)), 2) + 
				Math.pow(Math.min(Math.abs(y - yPos), model.worldLength - 
						Math.abs(y - yPos)), 2));
		coefficient = (double)(1 - distance/worldDiagonal);
		return coefficient;
	}

	/**
	 * Returns its knowledge base.
	 * 
	 * @return Returns its knowledge base.
	 */
	public KnowledgeBase getKnowledgeBase() {
		return myKnowledge;
	}
}
