
/*!
 * The imitating firms. Imitator is a child of Firm.
 * Imitators can only make vicarious exploration and vicarious exploitation.
 * Action is taken in two steps. First, rivals are identified. Secondly, 
 * knowledge is changed.
 * 
 *  \author Sirio Capizzi
 *  \author Guido Fioretti
 *  \author Ruggero Rossi
 *  \copyright GNU Public License.
 */
public class Imitator extends Firm {
	
	/**
	 * 
	 * @param model
	 * @param id
	 * @param s
	 * @param x
	 * @param y
	 * @param c
	 */
	public Imitator(Model model, int id, double s, int x, int y, int c) {
		super(model,id,s,x,y,c);
	}

	/**
	  * The following methods are called by updateKnowledge in Agent.
	  * They implement the actions carried out by Imitators:
	  * - experientialExploration (void);
	  * - experientialExploitation (void);
	  * - vicariousExploration, which makes use of:
	  *   - exitFieldInAnyCase;
	  * - vicariousExploitation;
	  * - exitField.
	 */
	
	/**
	 * Experiential Exploration (creation of a novel knowledge field).
	 */
	public void experientialExploration(KnowledgeField kf, double p, 
			boolean anyLowIntermediateCD) {}
	
	/**
	 * Experiential Exploitation (one's own knowledge field is deepened, 
	 * independently of rivals' knowledge).
	 */
	public void experientialExploitation(KnowledgeField kf, 
			double currentExpertise, double p, boolean anyLowIntermediateCD) {}
	
	
	/** 
	 * Vicarious Exploration (a knowledge field is copied from rivals, that is 
	 * new for the agent).
	 */
	public void vicariousExploration(KnowledgeField kf2, double wd2, 
			boolean anyIntermediateCD) {
		// Identify the kf with lowest expertise in my knowledge base.
		double lowestExpertise = myKnowledge.getLowestExpertise();
		KnowledgeField worstkf = myKnowledge.getKFwithLowestExpertise();

		// The conditions for vicarious exploration begin.
		if (anyIntermediateCD == true  	// cognitive distance is intermediate at least once
				&& size > 0  // there are sufficient financial resources
				&& !myKnowledge.isThereField(kf2)  // this firm lacks kf2
				&& (lowestExpertise < wd2 || myKnowledge.getNumFields() // the expertise of my
						< maxNumFields) // lowest-expertise field is lower than the expertise of kf2
				&& (!kf2.entailsUndefinedBEI() || // either kf2 does not entail any undefined aspect or
						kf2.entailsUndefinedBEI() && worstkf != null) // a non-null worst kf was found
				) {
			// If kf2 entails undefined aspects, these are substituted with the regular elements of 
			// the worst knowledge field, which is subsequently destroyed.
			if (kf2.entailsUndefinedBEI()) {
				kf2.instantiateWith(worstkf);
			}
			// If kf2 now entails regular aspects only, then vicarious exploration is carried out.
			if (!kf2.entailsUndefinedBEI()) {
				// The field with lowest expertise is eliminated if there are too many of them.
				if (myKnowledge.getNumFields() >= maxNumFields)
					exitFieldInAnyCase(worstkf);
				// The rivals' knowledge field is added to my knowledge.
				myKnowledge.setField(kf2, wd2);
				// Since in this case a new knowledge field is created, Common Knowledge is informed.
				commonKnowledge.updateField(kf2, 1);
				// The number of vicarious explorations is updated.
				vicExploration++;
				// Records actions to be displayed on graphs.
				for(Firm i: rivalsList)
					model.firms_actions.add(new FirmAction(this, i, 
							FirmAction.action_exploration, model.step));
				for(Firm i: rivalsList)
					model.InsertHistoricAction(this, i, 
							FirmAction.action_exploration);
			}
		}
	}
	
	/**
	 *  Vicarious Exploitation (if the rivals have deeper knowledge on a 
	 *  particular field, and if this knowledge intersects with a field of this 
	 *  agent, this knowledge is copied).
	 */
	public void vicariousExploitation(KnowledgeField kf1, double d1, double d2, double cd) {
		if (cd <= maxCriticalCognitiveDistance  // cognitive distance is low or 
												// intermediate
				&& d2 > d1  // the knowledge to be copied is deeper than the one 
							// of the calling firm
				&& size > 0  // there are sufficient financial resources
				) {
				// Expertise is absorbed to the extent it is similar.
			double additionalExpertise = (1 - cd);
			myKnowledge.setField(kf1, d1 + additionalExpertise);
			// The number of vicarious exploitations is updated
			vicExploitation++;
			// Records actions to be displayed on graphs.
			for(Firm i: rivalsList)
				model.firms_actions.add(new FirmAction(this, i, 
						FirmAction.action_exploitation, model.step));
			for(Firm i: rivalsList)
				model.InsertHistoricAction(this, i, 
						FirmAction.action_exploitation);
		}
	}
	
	/** 
	 * Exit a knowledge field if either its depth is zero, or it only entails 
	 * sundefined business elements.
	 */
	public void exitField(KnowledgeField kf) {
		if (myKnowledge.getExpertise(kf) <= 0) {
			myKnowledge.removeField(kf);  
				// the field is removed from my knowledge
			commonKnowledge.updateField(kf, -1);  // one less agent has kf		
			// The number of exits is updated.
			exit++;
		}
	}
	
	/**
	 *  Exit a knowledge field in any case (called by vicarious exploration)
	 * @param kf
	 */
	public void exitFieldInAnyCase(KnowledgeField kf) {
		myKnowledge.removeField(kf);  // the field is removed from my knowledge
		commonKnowledge.updateField(kf, -1);  // one less agent has kf		
		// The number of exits is updated.
		exit++;
	}
	
	/**
	 * These methods return an Imitator's actions.
	 */
	/**
	 * 
	 */
	public int getExperientialExploration() {
		return 0;
	}
	
	/**
	 * 
	 */
	public int getExperientialExploitation() {
		return 0;
	}
	
	/**
	 * 
	 */
	public int getVicariousExploration() {
		return vicExploration;
	}
	
	/**
	 * 
	 */
	public int getVicariousExploitation() {
		return vicExploitation;
	}
	
	/**
	 * 
	 */
	public int getExitField() {
		return exit;
	}
  
	/**
	 * This method communicates the color of a firm: Innovators are black, 
	 * Imitators are gray.
	 * This method is used in order to draw firms on networks.
	 */
	public java.awt.Color getColor() {
		return java.awt.Color.gray;
	}
}
