
import java.util.HashSet;
import java.util.Random;


/*!
 * The innovating firms. Innovator is a child of Firm.
 * Innovators are able to make both vicarious and experiential exploration, as 
 * well as both vicarious and experiential exploitation.
 * Action is taken in two steps. First, rivals are identified. 
 * Secondly, knowledge is changed.
 * 
 *  \author Sirio Capizzi
 *  \author Guido Fioretti
 *  \author Ruggero Rossi
 *  \copyright GNU Public License.
 */
public class Innovator extends Firm {
	
	/**
	 * 
	 * @param model
	 * @param id
	 * @param s
	 * @param x
	 * @param y
	 * @param c
	 */
	public Innovator(Model model, int id, double s, int x, int y, int c) {
		super(model,id,s,x,y,c);
	}
	
	/**
	  * The following methods are called by updateKnowledge in Firm.
	  * They implement the actions carried out by innovating firms:
	  * - experientialExploration, which makes use of:
	  *   - createNewField;
	  *   - exitFieldInAnyCase;
	  * - vicariousExploration, which makes use of:
	  *   - exitFieldInAnyCase;
	  * - experientialExploitation;
	  * - vicariousExploitation;
	  * - exitField.
	 */
	
	/**
	 *  Experiential Exploration (creation of a novel knowledge field).
	 */
	public void experientialExploration(KnowledgeField kf, double p, 
			boolean anyLowIntermediateCD) {
		// needs the subsequent method to create a new field
		
		Random rng = new Random(model.seed);  // the random numbers generator
		// The conditions for experiential exploration begin.
		if (anyLowIntermediateCD == false  	// cognitive distance is always high, 
											// or there are no rivals at all
				&& p > 0 && p <= minCriticalPerformance // and performance is 
											// low and not obtained by an 
											// isolated knowledge field
				&& myKnowledge.isThereField(kf)  // the knowledge field from 
											// which a novel knowledge field 
											// should be originated, is there
				&& size > 0) {  // there are sufficient financial resources
			// A new field is created starting from kf.
			KnowledgeField newkf = createNewField(kf);  // creates a new field
			double newExpertise = myKnowledge.getExpertise(kf)*rng.nextFloat();
			myKnowledge.setField(newkf, newExpertise);
			// Its existence is communicated to commonKnowledge. The number of 
			// experiential explorations is counted.
			commonKnowledge.updateField(newkf, 1);
			// The number of experiential explorations is updated.
			expExploration++;
			// The shallowest field is destroyed if the maximum number of fields 
			// has already been attained.
			if (myKnowledge.getNumFields() > maxNumFields) {
				KnowledgeField shallowestKF = 
						myKnowledge.getKFwithLowestExpertise();
				exitFieldInAnyCase(shallowestKF);
			}
			// Records actions to be displayed on networks.
			model.firms_actions.add(new FirmAction(this, this, 
					FirmAction.action_exploration, model.step));
			model.InsertHistoricAction(this, this, FirmAction.action_exploration);
		}
	}
	
	/**
	 *  This method is needed by experientialExploration.
	 *  It is the only place where new instances of a business element are 
	 *  created, beyond initialization.
	 * 
	 * @param src
	 * @return
	 */
	protected KnowledgeField createNewField(KnowledgeField src) {
		Random rng = new Random(model.seed);  // the random numbers generator
		// The number of new business elements is given by a parameter, but 
		// their choice is random.
		HashSet<Integer> chosen = new HashSet<Integer>();	// the business 
												// elements chosen to generate 
												// newkf by changing src
		KnowledgeField newkf=(KnowledgeField)src.cloneWithNewID();  
			// initially, newkf is just like the starting one
		for (int i=0; i<Model.newBusinessElements; i++) {  
			// for all BEIs to be substituted
			int whichBE=-1;
			do {
				whichBE = rng.nextInt(Model.numBusinessElements);
			}
			while (chosen.contains(whichBE));
			chosen.add(whichBE);
			BusinessElementInstance bei=new BusinessElementInstance(whichBE, false);
			Model.instances[whichBE].add(bei);
			newkf.substituteBEI(whichBE,bei);
		}
		return newkf;
	}
	
	/**
	 *  Experiential Exploitation (the expertise of one's own knowledge field 
	 *  increases independently of rivals' knowledge).
	 */
	public void experientialExploitation(KnowledgeField kf, 
			double currentExpertise, double p, boolean anyLowIntermediateCD) {
		if (anyLowIntermediateCD == false  	// cognitive distance is always high, 
											// or there are no rivals at all
				&& p > 0 && p > minCriticalPerformance  // and performance is 
											// above minimum and not obtained by 
											// an isolated field
				&& size > 0) {  // there are sufficient financial resources
			// Adds a depth that balances the natural decay (a fraction 
			// 'decayRate' of the maximum depth ever attained).
			myKnowledge.tuneFieldExpertise(kf, decayRate);
			// The number of experiential exploitations is updated.
			expExploitation++;
			// Records actions to be displayed on networks.
			model.firms_actions.add(new FirmAction(this, this, 
					FirmAction.action_exploitation, model.step));
			model.InsertHistoricAction(this, this, 
					FirmAction.action_exploitation);
		}
	}
	
	
	/**
	 *  Vicarious Exploration (a knowledge field is copied from rivals, that is 
	 *  new for the agent).
	 */
	public void vicariousExploration(KnowledgeField kf2, double wd2, 
			boolean anyIntermediateCD) {
		// Identify the kf with lowest expertise in my knowledge base.
		double lowestExpertise = myKnowledge.getLowestExpertise();
		KnowledgeField worstkf = myKnowledge.getKFwithLowestExpertise();

		// The conditions for vicarious exploration begin.
		if (anyIntermediateCD == true  	// cognitive distance is intermediate at least once
				&& size > 0  // there are sufficient financial resources
				&& !myKnowledge.isThereField(kf2)  // this firm lacks kf2
				&& (lowestExpertise < wd2 || myKnowledge.getNumFields() // the expertise of my
						< maxNumFields) // lowest-expertise field is lower than the expertise of kf2
				&& (!kf2.entailsUndefinedBEI() || // either kf2 does not entail any undefined aspect or
						kf2.entailsUndefinedBEI() && worstkf != null) // a non-null worst kf was found
				) {
			// If kf2 entails undefined aspects, these are substituted with the regular elements of 
			// the worst knowledge field, which is subsequently destroyed.
			if (kf2.entailsUndefinedBEI()) {
				kf2.instantiateWith(worstkf);
			}
			// If kf2 now entails regular aspects only, then vicarious exploration is carried out.
			if (!kf2.entailsUndefinedBEI()) {
				// The field with lowest depth is eliminated if there are too many of them.
				if (myKnowledge.getNumFields() >= maxNumFields)
					exitFieldInAnyCase(worstkf);
				// The rivals' knowledge field is added to my knowledge.
				myKnowledge.setField(kf2, wd2);
				// Since in this case a new knowledge field is created, Common Knowledge is informed.
				commonKnowledge.updateField(kf2, 1);
				// The number of vicarious explorations is updated.
				vicExploration++;
				// Records actions to be displayed on graphs.
				for(Firm i: rivalsList)
					model.firms_actions.add(new FirmAction(this, i, 
							FirmAction.action_exploration, model.step));
				for(Firm i: rivalsList)
					model.InsertHistoricAction(this, i, 
							FirmAction.action_exploration);
			}
		}
	}
	
	/** 
	 * Vicarious Exploitation (if the rivals have more expertise knowledge on a 
	 * particular field, and if this knowledge intersects with a field of this 
	 * agent, this knowledge is copied).
	 */
	public void vicariousExploitation(KnowledgeField kf1, double d1, double d2, double cd) {		
		if (cd <= maxCriticalCognitiveDistance  // cognitive distance is low or 
												// intermediate
				&& d2 > d1  // the knowledge field to be copied has more 
							// expertise than the one of the calling firm
				&& size > 0  // there are sufficient financial resources
				) {
				// Expertise is absorbed to the extent it is similar.
			double additionalExpertise = (1 - cd);
			myKnowledge.setField(kf1, d1 + additionalExpertise);
			// The number of vicarious exploitations is updated.
			vicExploitation++;
			// Records actions to be displayed on networks.
			for(Firm i: rivalsList)
				model.firms_actions.add(new FirmAction(this, i, 
						FirmAction.action_exploitation, model.step));
			for(Firm i: rivalsList)
				model.InsertHistoricAction(this, i, 
						FirmAction.action_exploitation);
		}
	}
	
	/**
	 *  Exit a knowledge field if either its depth is zero, or it only entails 
	 *  undefined business elements.
	 */
	public void exitField(KnowledgeField kf) {
		if (myKnowledge.getExpertise(kf) <= 0) {
			myKnowledge.removeField(kf);  	// the field is removed from my
											// knowledge
			commonKnowledge.updateField(kf, -1);  // one less agent has kf		
			// The number of exits is updated.
			exit++;
		}
	}
	
	/**
	 *  Exit a knowledge field in any case (called by experiential and 
	 *  vicarious exploration)
	 * @param kf
	 */
	public void exitFieldInAnyCase(KnowledgeField kf) {
		myKnowledge.removeField(kf);  // the field is removed from my knowledge
		commonKnowledge.updateField(kf, -1);  // one less agent has kf		
		// The number of exits is updated.
		exit++;		
	}
	
	/**
	 * These methods return an Innovator's actions.
	 */
	/**
	 * 
	 */
	public int getExperientialExploration() {
		return expExploration;
	}
	
	/**
	 * 
	 */
	public int getExperientialExploitation() {
		return expExploitation;
	}
	
	/**
	 * 
	 */
	public int getVicariousExploration() {
		return vicExploration;
	}
	
	/**
	 * 
	 */
	public int getVicariousExploitation() {
		return vicExploitation;
	}
	
	/**
	 * 
	 */
	public int getExitField() {
		return exit;
	}
  
	/**
	 * This method communicates the color of a firm: Innovators are black, 
	 * Imitators are gray.
	 * This method is used in order to draw firms on networks.
	 */
	public java.awt.Color getColor() {
		return java.awt.Color.black;
	}
}
